# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Tools for running presubmit checks in a Git repository.

Presubmit checks are defined as a function or other callable. The function may
take either no arguments or a list of the paths on which to run. Presubmit
checks communicate failure by raising any exception.

For example, either of these functions may be used as presubmit checks:

  @pw_presubmit.filter_paths(endswith='.py')
  def file_contains_ni(ctx: PresubmitContext):
      for path in ctx.paths:
          with open(path) as file:
              contents = file.read()
              if 'ni' not in contents and 'nee' not in contents:
                  raise PresumitFailure('Files must say "ni"!', path=path)

  def run_the_build():
      subprocess.run(['make', 'release'], check=True)

Presubmit checks that accept a list of paths may use the filter_paths decorator
to automatically filter the paths list for file types they care about. See the
pragma_once function for an example.

See pigweed_presbumit.py for an example of how to define presubmit checks.
"""

from __future__ import annotations

import collections
import contextlib
import copy
import dataclasses
import enum
from inspect import Parameter, signature
import itertools
import json
import logging
import os
from pathlib import Path
import re
import shutil
import signal
import subprocess
import sys
import tempfile as tf
import time
import types
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Pattern,
    Sequence,
    Set,
    Tuple,
    Union,
)
import urllib

import pw_cli.color
import pw_cli.env
import pw_env_setup.config_file
from pw_package import package_manager
from pw_presubmit import git_repo, tools
from pw_presubmit.tools import plural

_LOG: logging.Logger = logging.getLogger(__name__)

_COLOR = pw_cli.color.colors()

_SUMMARY_BOX = '══╦╗ ║║══╩╝'
_CHECK_UPPER = '━━━┓       '
_CHECK_LOWER = '       ━━━┛'

WIDTH = 80

_LEFT = 7
_RIGHT = 11


def _title(msg, style=_SUMMARY_BOX) -> str:
    msg = f' {msg} '.center(WIDTH - 2)
    return tools.make_box('^').format(*style, section1=msg, width1=len(msg))


def _format_time(time_s: float) -> str:
    minutes, seconds = divmod(time_s, 60)
    if minutes < 60:
        return f' {int(minutes)}:{seconds:04.1f}'
    hours, minutes = divmod(minutes, 60)
    return f'{int(hours):d}:{int(minutes):02}:{int(seconds):02}'


def _box(style, left, middle, right, box=tools.make_box('><>')) -> str:
    return box.format(
        *style,
        section1=left + ('' if left.endswith(' ') else ' '),
        width1=_LEFT,
        section2=' ' + middle,
        width2=WIDTH - _LEFT - _RIGHT - 4,
        section3=right + ' ',
        width3=_RIGHT,
    )


class PresubmitFailure(Exception):
    """Optional exception to use for presubmit failures."""

    def __init__(
        self,
        description: str = '',
        path: Optional[Path] = None,
        line: Optional[int] = None,
    ):
        line_part: str = ''
        if line is not None:
            line_part = f'{line}:'
        super().__init__(
            f'{path}:{line_part} {description}' if path else description
        )


class PresubmitResult(enum.Enum):
    PASS = 'PASSED'  # Check completed successfully.
    FAIL = 'FAILED'  # Check failed.
    CANCEL = 'CANCEL'  # Check didn't complete.

    def colorized(self, width: int, invert: bool = False) -> str:
        if self is PresubmitResult.PASS:
            color = _COLOR.black_on_green if invert else _COLOR.green
        elif self is PresubmitResult.FAIL:
            color = _COLOR.black_on_red if invert else _COLOR.red
        elif self is PresubmitResult.CANCEL:
            color = _COLOR.yellow
        else:
            color = lambda value: value

        padding = (width - len(self.value)) // 2 * ' '
        return padding + color(self.value) + padding


class Program(collections.abc.Sequence):
    """A sequence of presubmit checks; basically a tuple with a name."""

    def __init__(self, name: str, steps: Iterable[Callable]):
        self.name = name

        def ensure_check(step):
            if isinstance(step, Check):
                return step
            return Check(step)

        self._steps: tuple[Check, ...] = tuple(
            {ensure_check(s): None for s in tools.flatten(steps)}
        )

    def __getitem__(self, i):
        return self._steps[i]

    def __len__(self):
        return len(self._steps)

    def __str__(self):
        return self.name

    def title(self):
        return f'{self.name if self.name else ""} presubmit checks'.strip()


class Programs(collections.abc.Mapping):
    """A mapping of presubmit check programs.

    Use is optional. Helpful when managing multiple presubmit check programs.
    """

    def __init__(self, **programs: Sequence):
        """Initializes a name: program mapping from the provided keyword args.

        A program is a sequence of presubmit check functions. The sequence may
        contain nested sequences, which are flattened.
        """
        self._programs: Dict[str, Program] = {
            name: Program(name, checks) for name, checks in programs.items()
        }

    def all_steps(self) -> Dict[str, Check]:
        return {c.name: c for c in itertools.chain(*self.values())}

    def __getitem__(self, item: str) -> Program:
        return self._programs[item]

    def __iter__(self) -> Iterator[str]:
        return iter(self._programs)

    def __len__(self) -> int:
        return len(self._programs)


@dataclasses.dataclass(frozen=True)
class FormatOptions:
    python_formatter: Optional[str] = 'yapf'
    black_path: Optional[str] = 'black'

    # TODO(b/264578594) Add exclude to pigweed.json file.
    # exclude: Sequence[re.Pattern] = dataclasses.field(default_factory=list)

    @staticmethod
    def load() -> 'FormatOptions':
        config = pw_env_setup.config_file.load()
        fmt = config.get('pw', {}).get('pw_presubmit', {}).get('format', {})
        return FormatOptions(
            python_formatter=fmt.get('python_formatter', 'yapf'),
            black_path=fmt.get('black_path', 'black'),
            # exclude=tuple(re.compile(x) for x in fmt.get('exclude', ())),
        )


@dataclasses.dataclass
class LuciPipeline:
    round: int
    builds_from_previous_iteration: Sequence[str]

    @staticmethod
    def create(
        bbid: int,
        fake_pipeline_props: Optional[Dict[str, Any]] = None,
    ) -> Optional['LuciPipeline']:
        pipeline_props: Dict[str, Any]
        if fake_pipeline_props is not None:
            pipeline_props = fake_pipeline_props
        else:
            pipeline_props = (
                get_buildbucket_info(bbid)
                .get('input', {})
                .get('properties', {})
                .get('$pigweed/pipeline', {})
            )
        if not pipeline_props.get('inside_a_pipeline', False):
            return None

        return LuciPipeline(
            round=int(pipeline_props['round']),
            builds_from_previous_iteration=list(
                pipeline_props['builds_from_previous_iteration']
            ),
        )


def get_buildbucket_info(bbid) -> Dict[str, Any]:
    if not bbid or not shutil.which('bb'):
        return {}

    output = subprocess.check_output(
        ['bb', 'get', '-json', '-p', f'{bbid}'], text=True
    )
    return json.loads(output)


def download_cas_artifact(
    ctx: PresubmitContext, digest: str, output_dir: str
) -> None:
    """Downloads the given digest to the given outputdirectory

    Args:
        ctx: the presubmit context
        digest:
        a string digest in the form "<digest hash>/<size bytes>"
        i.e 693a04e41374150d9d4b645fccb49d6f96e10b527c7a24b1e17b331f508aa73b/86
        output_dir: the directory we want to download the artifacts to
    """
    if ctx.luci is None:
        raise PresubmitFailure('Lucicontext is None')
    cmd = [
        'cas',
        'download',
        '-cas-instance',
        ctx.luci.cas_instance,
        '-digest',
        digest,
        '-dir',
        output_dir,
    ]
    try:
        subprocess.check_call(cmd)
    except subprocess.CalledProcessError as failure:
        raise PresubmitFailure('cas download failed') from failure


def archive_cas_artifact(
    ctx: PresubmitContext, root: str, upload_paths: List[str]
) -> str:
    """Uploads the given artifacts into cas

    Args:
        ctx: the presubmit context
        root: root directory of archived tree, should be absolutepath.
        paths: path to archived files/dirs, should be absolute path.
            If empty, [root] will be used.

    Returns:
        A string digest in the form "<digest hash>/<size bytes>"
        i.e 693a04e41374150d9d4b645fccb49d6f96e10b527c7a24b1e17b331f508aa73b/86
    """
    if ctx.luci is None:
        raise PresubmitFailure('Lucicontext is None')
    assert os.path.abspath(root)
    if not upload_paths:
        upload_paths = [root]
    for path in upload_paths:
        assert os.path.abspath(path)

    with tf.NamedTemporaryFile(mode='w+t') as tmp_digest_file:
        with tf.NamedTemporaryFile(mode='w+t') as tmp_paths_file:
            json_paths = json.dumps(
                [
                    [str(root), str(os.path.relpath(path, root))]
                    for path in upload_paths
                ]
            )
            tmp_paths_file.write(json_paths)
            tmp_paths_file.seek(0)
            cmd = [
                'cas',
                'archive',
                '-cas-instance',
                ctx.luci.cas_instance,
                '-paths-json',
                tmp_paths_file.name,
                '-dump-digest',
                tmp_digest_file.name,
            ]
            try:
                subprocess.check_call(cmd)
            except subprocess.CalledProcessError as failure:
                raise PresubmitFailure('cas archive failed') from failure

            tmp_digest_file.seek(0)
            uploaded_digest = tmp_digest_file.read()
            return uploaded_digest


@dataclasses.dataclass
class LuciTrigger:
    """Details the pending change or submitted commit triggering the build."""

    number: int
    remote: str
    branch: str
    ref: str
    gerrit_name: str
    submitted: bool

    @property
    def gerrit_url(self):
        if not self.number:
            return self.gitiles_url
        return 'https://{}-review.googlesource.com/c/{}'.format(
            self.gerrit_name, self.number
        )

    @property
    def gitiles_url(self):
        return '{}/+/{}'.format(self.remote, self.ref)

    @staticmethod
    def create_from_environment(
        env: Optional[Dict[str, str]] = None,
    ) -> Sequence['LuciTrigger']:
        if not env:
            env = os.environ.copy()
        raw_path = env.get('TRIGGERING_CHANGES_JSON')
        if not raw_path:
            return ()
        path = Path(raw_path)
        if not path.is_file():
            return ()

        result = []
        with open(path, 'r') as ins:
            for trigger in json.load(ins):
                keys = {
                    'number',
                    'remote',
                    'branch',
                    'ref',
                    'gerrit_name',
                    'submitted',
                }
                if keys <= trigger.keys():
                    result.append(LuciTrigger(**{x: trigger[x] for x in keys}))

        return tuple(result)

    @staticmethod
    def create_for_testing():
        change = {
            'number': 123456,
            'remote': 'https://pigweed.googlesource.com/pigweed/pigweed',
            'branch': 'main',
            'ref': 'refs/changes/56/123456/1',
            'gerrit_name': 'pigweed',
            'submitted': True,
        }
        with tf.TemporaryDirectory() as tempdir:
            changes_json = Path(tempdir) / 'changes.json'
            with changes_json.open('w') as outs:
                json.dump([change], outs)
            env = {'TRIGGERING_CHANGES_JSON': changes_json}
            return LuciTrigger.create_from_environment(env)


@dataclasses.dataclass
class LuciContext:
    """LUCI-specific information about the environment."""

    buildbucket_id: int
    build_number: int
    project: str
    bucket: str
    builder: str
    swarming_server: str
    swarming_task_id: str
    cas_instance: str
    pipeline: Optional[LuciPipeline]
    triggers: Sequence[LuciTrigger] = dataclasses.field(default_factory=tuple)

    @staticmethod
    def create_from_environment(
        env: Optional[Dict[str, str]] = None,
        fake_pipeline_props: Optional[Dict[str, Any]] = None,
    ) -> Optional['LuciContext']:
        """Create a LuciContext from the environment."""

        if not env:
            env = os.environ.copy()

        luci_vars = [
            'BUILDBUCKET_ID',
            'BUILDBUCKET_NAME',
            'BUILD_NUMBER',
            'SWARMING_TASK_ID',
            'SWARMING_SERVER',
        ]
        if any(x for x in luci_vars if x not in env):
            return None

        project, bucket, builder = env['BUILDBUCKET_NAME'].split(':')

        bbid: int = 0
        pipeline: Optional[LuciPipeline] = None
        try:
            bbid = int(env['BUILDBUCKET_ID'])
            pipeline = LuciPipeline.create(bbid, fake_pipeline_props)

        except ValueError:
            pass

        # Logic to identify cas instance from swarming server is derived from
        # https://chromium.googlesource.com/infra/luci/recipes-py/+/main/recipe_modules/cas/api.py
        swarm_server = env['SWARMING_SERVER']
        cas_project = urllib.parse.urlparse(swarm_server).netloc.split('.')[0]
        cas_instance = f'projects/{cas_project}/instances/default_instance'

        result = LuciContext(
            buildbucket_id=bbid,
            build_number=int(env['BUILD_NUMBER']),
            project=project,
            bucket=bucket,
            builder=builder,
            swarming_server=env['SWARMING_SERVER'],
            swarming_task_id=env['SWARMING_TASK_ID'],
            cas_instance=cas_instance,
            pipeline=pipeline,
            triggers=LuciTrigger.create_from_environment(env),
        )
        _LOG.debug('%r', result)
        return result

    @staticmethod
    def create_for_testing():
        env = {
            'BUILDBUCKET_ID': '881234567890',
            'BUILDBUCKET_NAME': 'pigweed:bucket.try:builder-name',
            'BUILD_NUMBER': '123',
            'SWARMING_SERVER': 'https://chromium-swarm.appspot.com',
            'SWARMING_TASK_ID': 'cd2dac62d2',
        }
        return LuciContext.create_from_environment(env, {})


@dataclasses.dataclass
class FormatContext:
    """Context passed into formatting helpers.

    This class is a subset of PresubmitContext containing only what's needed by
    formatters.

    For full documentation on the members see the PresubmitContext section of
    pw_presubmit/docs.rst.

    Args:
        root: Source checkout root directory
        output_dir: Output directory for this specific language
        paths: Modified files for the presubmit step to check (often used in
            formatting steps but ignored in compile steps)
        package_root: Root directory for pw package installations
        format_options: Formatting options, derived from pigweed.json
    """

    root: Optional[Path]
    output_dir: Path
    paths: Tuple[Path, ...]
    package_root: Path
    format_options: FormatOptions


@dataclasses.dataclass
class PresubmitContext:  # pylint: disable=too-many-instance-attributes
    """Context passed into presubmit checks.

    For full documentation on the members see pw_presubmit/docs.rst.

    Args:
        root: Source checkout root directory
        repos: Repositories (top-level and submodules) processed by
            pw presubmit
        output_dir: Output directory for this specific presubmit step
        failure_summary_log: Path where steps should write a brief summary of
            any failures encountered for use by other tooling.
        paths: Modified files for the presubmit step to check (often used in
            formatting steps but ignored in compile steps)
        all_paths: All files in the tree.
        package_root: Root directory for pw package installations
        override_gn_args: Additional GN args processed by build.gn_gen()
        luci: Information about the LUCI build or None if not running in LUCI
        format_options: Formatting options, derived from pigweed.json
        num_jobs: Number of jobs to run in parallel
        continue_after_build_error: For steps that compile, don't exit on the
            first compilation error
    """

    root: Path
    repos: Tuple[Path, ...]
    output_dir: Path
    failure_summary_log: Path
    paths: Tuple[Path, ...]
    all_paths: Tuple[Path, ...]
    package_root: Path
    luci: Optional[LuciContext]
    override_gn_args: Dict[str, str]
    format_options: FormatOptions
    num_jobs: Optional[int] = None
    continue_after_build_error: bool = False
    _failed: bool = False

    @property
    def failed(self) -> bool:
        return self._failed

    def fail(
        self,
        description: str,
        path: Optional[Path] = None,
        line: Optional[int] = None,
    ):
        """Add a failure to this presubmit step.

        If this is called at least once the step fails, but not immediately—the
        check is free to continue and possibly call this method again.
        """
        _LOG.warning('%s', PresubmitFailure(description, path, line))
        self._failed = True

    @staticmethod
    def create_for_testing():
        parsed_env = pw_cli.env.pigweed_environment()
        root = parsed_env.PW_PROJECT_ROOT
        presubmit_root = root / 'out' / 'presubmit'
        return PresubmitContext(
            root=root,
            repos=(root,),
            output_dir=presubmit_root / 'test',
            failure_summary_log=presubmit_root / 'failure-summary.log',
            paths=(root / 'foo.cc', root / 'foo.py'),
            all_paths=(root / 'BUILD.gn', root / 'foo.cc', root / 'foo.py'),
            package_root=root / 'environment' / 'packages',
            luci=None,
            override_gn_args={},
            format_options=FormatOptions(),
        )


class FileFilter:
    """Allows checking if a path matches a series of filters.

    Positive filters (e.g. the file name matches a regex) and negative filters
    (path does not match a regular expression) may be applied.
    """

    _StrOrPattern = Union[Pattern, str]

    def __init__(
        self,
        *,
        exclude: Iterable[_StrOrPattern] = (),
        endswith: Iterable[str] = (),
        name: Iterable[_StrOrPattern] = (),
        suffix: Iterable[str] = (),
    ) -> None:
        """Creates a FileFilter with the provided filters.

        Args:
            endswith: True if the end of the path is equal to any of the passed
                      strings
            exclude: If any of the passed regular expresion match return False.
                     This overrides and other matches.
            name: Regexs to match with file names(pathlib.Path.name). True if
                  the resulting regex matches the entire file name.
            suffix: True if final suffix (as determined by pathlib.Path) is
                    matched by any of the passed str.
        """
        self.exclude = tuple(re.compile(i) for i in exclude)

        self.endswith = tuple(endswith)
        self.name = tuple(re.compile(i) for i in name)
        self.suffix = tuple(suffix)

    def matches(self, path: Union[str, Path]) -> bool:
        """Returns true if the path matches any filter but not an exclude.

        If no positive filters are specified, any paths that do not match a
        negative filter are considered to match.

        If 'path' is a Path object it is rendered as a posix path (i.e.
        using "/" as the path seperator) before testing with 'exclude' and
        'endswith'.
        """

        posix_path = path.as_posix() if isinstance(path, Path) else path
        if any(bool(exp.search(posix_path)) for exp in self.exclude):
            return False

        # If there are no positive filters set, accept all paths.
        no_filters = not self.endswith and not self.name and not self.suffix

        path_obj = Path(path)
        return (
            no_filters
            or path_obj.suffix in self.suffix
            or any(regex.fullmatch(path_obj.name) for regex in self.name)
            or any(posix_path.endswith(end) for end in self.endswith)
        )

    def filter(self, paths: Sequence[Union[str, Path]]) -> Sequence[Path]:
        return [Path(x) for x in paths if self.matches(x)]

    def apply_to_check(self, always_run: bool = False) -> Callable:
        def wrapper(func: Callable) -> Check:
            if isinstance(func, Check):
                clone = copy.copy(func)
                clone.filter = self
                clone.always_run = clone.always_run or always_run
                return clone

            return Check(check=func, path_filter=self, always_run=always_run)

        return wrapper


def _print_ui(*args) -> None:
    """Prints to stdout and flushes to stay in sync with logs on stderr."""
    print(*args, flush=True)


@dataclasses.dataclass
class FilteredCheck:
    check: Check
    paths: Sequence[Path]
    substep: Optional[str] = None

    @property
    def name(self) -> str:
        return self.check.name

    def run(self, ctx: PresubmitContext, count: int, total: int):
        return self.check.run(ctx, count, total, self.substep)


class Presubmit:
    """Runs a series of presubmit checks on a list of files."""

    def __init__(
        self,
        root: Path,
        repos: Sequence[Path],
        output_directory: Path,
        paths: Sequence[Path],
        all_paths: Sequence[Path],
        package_root: Path,
        override_gn_args: Dict[str, str],
        continue_after_build_error: bool,
    ):
        self._root = root.resolve()
        self._repos = tuple(repos)
        self._output_directory = output_directory.resolve()
        self._paths = tuple(paths)
        self._all_paths = tuple(all_paths)
        self._relative_paths = tuple(
            tools.relative_paths(self._paths, self._root)
        )
        self._package_root = package_root.resolve()
        self._override_gn_args = override_gn_args
        self._continue_after_build_error = continue_after_build_error

    def run(
        self,
        program: Program,
        keep_going: bool = False,
        substep: Optional[str] = None,
    ) -> bool:
        """Executes a series of presubmit checks on the paths."""

        checks = self.apply_filters(program)
        if substep:
            assert (
                len(checks) == 1
            ), 'substeps not supported with multiple steps'
            checks[0].substep = substep

        _LOG.debug('Running %s for %s', program.title(), self._root.name)
        _print_ui(_title(f'{self._root.name}: {program.title()}'))

        _LOG.info(
            '%d of %d checks apply to %s in %s',
            len(checks),
            len(program),
            plural(self._paths, 'file'),
            self._root,
        )

        _print_ui()
        for line in tools.file_summary(self._relative_paths):
            _print_ui(line)
        _print_ui()

        if not self._paths:
            _print_ui(_COLOR.yellow('No files are being checked!'))

        _LOG.debug('Checks:\n%s', '\n'.join(c.name for c in checks))

        start_time: float = time.time()
        passed, failed, skipped = self._execute_checks(checks, keep_going)
        self._log_summary(time.time() - start_time, passed, failed, skipped)

        return not failed and not skipped

    def apply_filters(self, program: Sequence[Callable]) -> List[FilteredCheck]:
        """Returns list of FilteredCheck for checks that should run."""
        checks = [c if isinstance(c, Check) else Check(c) for c in program]
        filter_to_checks: Dict[
            FileFilter, List[Check]
        ] = collections.defaultdict(list)

        for chk in checks:
            filter_to_checks[chk.filter].append(chk)

        check_to_paths = self._map_checks_to_paths(filter_to_checks)
        return [
            FilteredCheck(c, check_to_paths[c])
            for c in checks
            if c in check_to_paths
        ]

    def _map_checks_to_paths(
        self, filter_to_checks: Dict[FileFilter, List[Check]]
    ) -> Dict[Check, Sequence[Path]]:
        checks_to_paths: Dict[Check, Sequence[Path]] = {}

        posix_paths = tuple(p.as_posix() for p in self._relative_paths)

        for filt, checks in filter_to_checks.items():
            filtered_paths = tuple(
                path
                for path, filter_path in zip(self._paths, posix_paths)
                if filt.matches(filter_path)
            )

            for chk in checks:
                if filtered_paths or chk.always_run:
                    checks_to_paths[chk] = filtered_paths
                else:
                    _LOG.debug('Skipping "%s": no relevant files', chk.name)

        return checks_to_paths

    def _log_summary(
        self, time_s: float, passed: int, failed: int, skipped: int
    ) -> None:
        summary_items = []
        if passed:
            summary_items.append(f'{passed} passed')
        if failed:
            summary_items.append(f'{failed} failed')
        if skipped:
            summary_items.append(f'{skipped} not run')
        summary = ', '.join(summary_items) or 'nothing was done'

        if failed or skipped:
            result = PresubmitResult.FAIL
        else:
            result = PresubmitResult.PASS
        total = passed + failed + skipped

        _LOG.debug(
            'Finished running %d checks on %s in %.1f s',
            total,
            plural(self._paths, 'file'),
            time_s,
        )
        _LOG.debug('Presubmit checks %s: %s', result.value, summary)

        _print_ui(
            _box(
                _SUMMARY_BOX,
                result.colorized(_LEFT, invert=True),
                f'{total} checks on {plural(self._paths, "file")}: {summary}',
                _format_time(time_s),
            )
        )

    def _create_presubmit_context(  # pylint: disable=no-self-use
        self, **kwargs
    ):
        """Create a PresubmitContext. Override if needed in subclasses."""
        return PresubmitContext(**kwargs)

    @contextlib.contextmanager
    def _context(self, filtered_check: FilteredCheck):
        # There are many characters banned from filenames on Windows. To
        # simplify things, just strip everything that's not a letter, digit,
        # or underscore.
        sanitized_name = re.sub(r'[\W_]+', '_', filtered_check.name).lower()
        output_directory = self._output_directory.joinpath(sanitized_name)
        os.makedirs(output_directory, exist_ok=True)

        failure_summary_log = output_directory / 'failure-summary.log'
        failure_summary_log.unlink(missing_ok=True)

        handler = logging.FileHandler(
            output_directory.joinpath('step.log'), mode='w'
        )
        handler.setLevel(logging.DEBUG)

        try:
            _LOG.addHandler(handler)

            yield self._create_presubmit_context(
                root=self._root,
                repos=self._repos,
                output_dir=output_directory,
                failure_summary_log=failure_summary_log,
                paths=filtered_check.paths,
                all_paths=self._all_paths,
                package_root=self._package_root,
                override_gn_args=self._override_gn_args,
                continue_after_build_error=self._continue_after_build_error,
                luci=LuciContext.create_from_environment(),
                format_options=FormatOptions.load(),
            )

        finally:
            _LOG.removeHandler(handler)

    def _execute_checks(
        self, program: List[FilteredCheck], keep_going: bool
    ) -> Tuple[int, int, int]:
        """Runs presubmit checks; returns (passed, failed, skipped) lists."""
        passed = failed = 0

        for i, filtered_check in enumerate(program, 1):
            with self._context(filtered_check) as ctx:
                result = filtered_check.run(ctx, i, len(program))

            if result is PresubmitResult.PASS:
                passed += 1
            elif result is PresubmitResult.CANCEL:
                break
            else:
                failed += 1
                if not keep_going:
                    break

        return passed, failed, len(program) - passed - failed


def _process_pathspecs(
    repos: Iterable[Path], pathspecs: Iterable[str]
) -> Dict[Path, List[str]]:
    pathspecs_by_repo: Dict[Path, List[str]] = {repo: [] for repo in repos}
    repos_with_paths: Set[Path] = set()

    for pathspec in pathspecs:
        # If the pathspec is a path to an existing file, only use it for the
        # repo it is in.
        if os.path.exists(pathspec):
            # Raise an exception if the path exists but is not in a known repo.
            repo = git_repo.within_repo(pathspec)
            if repo not in pathspecs_by_repo:
                raise ValueError(
                    f'{pathspec} is not in a Git repository in this presubmit'
                )

            # Make the path relative to the repo's root.
            pathspecs_by_repo[repo].append(os.path.relpath(pathspec, repo))
            repos_with_paths.add(repo)
        else:
            # Pathspecs that are not paths (e.g. '*.h') are used for all repos.
            for patterns in pathspecs_by_repo.values():
                patterns.append(pathspec)

    # If any paths were specified, only search for paths in those repos.
    if repos_with_paths:
        for repo in set(pathspecs_by_repo) - repos_with_paths:
            del pathspecs_by_repo[repo]

    return pathspecs_by_repo


def run(  # pylint: disable=too-many-arguments,too-many-locals
    program: Sequence[Check],
    root: Path,
    repos: Collection[Path] = (),
    base: Optional[str] = None,
    paths: Sequence[str] = (),
    exclude: Sequence[Pattern] = (),
    output_directory: Optional[Path] = None,
    package_root: Optional[Path] = None,
    only_list_steps: bool = False,
    override_gn_args: Sequence[Tuple[str, str]] = (),
    keep_going: bool = False,
    continue_after_build_error: bool = False,
    presubmit_class: type = Presubmit,
    list_steps_file: Optional[Path] = None,
    substep: Optional[str] = None,
) -> bool:
    """Lists files in the current Git repo and runs a Presubmit with them.

    This changes the directory to the root of the Git repository after listing
    paths, so all presubmit checks can assume they run from there.

    The paths argument contains Git pathspecs. If no pathspecs are provided, all
    paths in all repos are included. If paths to files or directories are
    provided, only files within those repositories are searched. Patterns are
    searched across all repositories. For example, if the pathspecs "my_module/"
    and "*.h", paths under "my_module/" in the containing repo and paths in all
    repos matching "*.h" will be included in the presubmit.

    Args:
        program: list of presubmit check functions to run
        root: root path of the project
        repos: paths to the roots of Git repositories to check
        name: name to use to refer to this presubmit check run
        base: optional base Git commit to list files against
        paths: optional list of Git pathspecs to run the checks against
        exclude: regular expressions for Posix-style paths to exclude
        output_directory: where to place output files
        package_root: where to place package files
        only_list_steps: print step names instead of running them
        override_gn_args: additional GN args to set on steps
        keep_going: continue running presubmit steps after a step fails
        continue_after_build_error: continue building if a build step fails
        presubmit_class: class to use to run Presubmits, should inherit from
            Presubmit class above
        list_steps_file: File created by --only-list-steps, used to keep from
            recalculating affected files.
        substep: run only part of a single check

    Returns:
        True if all presubmit checks succeeded
    """
    repos = [repo.resolve() for repo in repos]

    non_empty_repos = []
    for repo in repos:
        if list(repo.iterdir()):
            non_empty_repos.append(repo)
            if git_repo.root(repo) != repo:
                raise ValueError(
                    f'{repo} is not the root of a Git repo; '
                    'presubmit checks must be run from a Git repo'
                )
    repos = non_empty_repos

    pathspecs_by_repo = _process_pathspecs(repos, paths)

    all_files: List[Path] = []
    modified_files: List[Path] = []
    list_steps_data: Dict[str, Any] = {}

    if list_steps_file:
        with list_steps_file.open() as ins:
            list_steps_data = json.load(ins)
        all_files.extend(list_steps_data['all_files'])
        for step in list_steps_data['steps']:
            modified_files.extend(Path(x) for x in step.get("paths", ()))
        modified_files = sorted(set(modified_files))
        _LOG.info(
            'Loaded %d paths from file %s',
            len(modified_files),
            list_steps_file,
        )

    else:
        for repo, pathspecs in pathspecs_by_repo.items():
            all_files_repo = tuple(
                tools.exclude_paths(
                    exclude, git_repo.list_files(None, pathspecs, repo), root
                )
            )
            all_files += all_files_repo

            if base is None:
                modified_files += all_files_repo
            else:
                modified_files += tools.exclude_paths(
                    exclude, git_repo.list_files(base, pathspecs, repo), root
                )

            _LOG.info(
                'Checking %s',
                git_repo.describe_files(
                    repo, repo, base, pathspecs, exclude, root
                ),
            )

    if output_directory is None:
        output_directory = root / '.presubmit'

    if package_root is None:
        package_root = output_directory / 'packages'

    presubmit = presubmit_class(
        root=root,
        repos=repos,
        output_directory=output_directory,
        paths=modified_files,
        all_paths=all_files,
        package_root=package_root,
        override_gn_args=dict(override_gn_args or {}),
        continue_after_build_error=continue_after_build_error,
    )

    if only_list_steps:
        steps: List[Dict] = []
        for filtered_check in presubmit.apply_filters(program):
            step = {
                'name': filtered_check.name,
                'paths': [str(x) for x in filtered_check.paths],
            }
            substeps = filtered_check.check.substeps()
            if len(substeps) > 1:
                step['substeps'] = [x.name for x in substeps]
            steps.append(step)

        list_steps_data = {
            'steps': steps,
            'all_files': [str(x) for x in all_files],
        }
        json.dump(list_steps_data, sys.stdout, indent=2)
        sys.stdout.write('\n')
        return True

    if not isinstance(program, Program):
        program = Program('', program)

    return presubmit.run(program, keep_going, substep=substep)


def _make_str_tuple(value: Union[Iterable[str], str]) -> Tuple[str, ...]:
    return tuple([value] if isinstance(value, str) else value)


def check(*args, **kwargs):
    """Turn a function into a presubmit check.

    Args:
        *args: Passed through to function.
        *kwargs: Passed through to function.

    If only one argument is provided and it's a function, this function acts
    as a decorator and creates a Check from the function. Example of this kind
    of usage:

    @check
    def pragma_once(ctx: PresubmitContext):
        pass

    Otherwise, save the arguments, and return a decorator that turns a function
    into a Check, but with the arguments added onto the Check constructor.
    Example of this kind of usage:

    @check(name='pragma_twice')
    def pragma_once(ctx: PresubmitContext):
        pass
    """
    if (
        len(args) == 1
        and isinstance(args[0], types.FunctionType)
        and not kwargs
    ):
        # Called as a regular decorator.
        return Check(args[0])

    def decorator(check_function):
        return Check(check_function, *args, **kwargs)

    return decorator


@dataclasses.dataclass
class SubStep:
    name: Optional[str]
    _func: Callable[..., PresubmitResult]
    args: Sequence[Any] = ()
    kwargs: Dict[str, Any] = dataclasses.field(default_factory=lambda: {})

    def __call__(self, ctx: PresubmitContext) -> PresubmitResult:
        if self.name:
            _LOG.info('%s', self.name)
        return self._func(ctx, *self.args, **self.kwargs)


class Check:
    """Wraps a presubmit check function.

    This class consolidates the logic for running and logging a presubmit check.
    It also supports filtering the paths passed to the presubmit check.
    """

    def __init__(
        self,
        check: Union[  # pylint: disable=redefined-outer-name
            Callable, Iterable[SubStep]
        ],
        path_filter: FileFilter = FileFilter(),
        always_run: bool = True,
        name: Optional[str] = None,
        doc: Optional[str] = None,
    ) -> None:
        # Since Check wraps a presubmit function, adopt that function's name.
        self.name: str = ''
        self.doc: str = ''
        if isinstance(check, Check):
            self.name = check.name
            self.doc = check.doc
        elif callable(check):
            self.name = check.__name__
            self.doc = check.__doc__ or ''

        if name:
            self.name = name
        if doc:
            self.doc = doc

        if not self.name:
            raise ValueError('no name for step')

        self._substeps_raw: Iterable[SubStep]
        if isinstance(check, collections.abc.Iterator):
            self._substeps_raw = check
        else:
            assert callable(check)
            _ensure_is_valid_presubmit_check_function(check)
            self._substeps_raw = iter((SubStep(None, check),))
        self._substeps_saved: Sequence[SubStep] = ()

        self.filter = path_filter
        self.always_run: bool = always_run

    def substeps(self) -> Sequence[SubStep]:
        """Return the SubSteps of the current step.

        This is where the list of SubSteps is actually evaluated. It can't be
        evaluated in the constructor because the Iterable passed into the
        constructor might not be ready yet.
        """
        if not self._substeps_saved:
            self._substeps_saved = tuple(self._substeps_raw)
        return self._substeps_saved

    def __repr__(self):
        # This returns just the name so it's easy to show the entire list of
        # steps with '--help'.
        return self.name

    def unfiltered(self) -> Check:
        """Create a new check identical to this one, but without the filter."""
        clone = copy.copy(self)
        clone.filter = FileFilter()
        return clone

    def with_filter(
        self,
        *,
        endswith: Iterable[str] = (),
        exclude: Iterable[Union[Pattern[str], str]] = (),
    ) -> Check:
        """Create a new check identical to this one, but with extra filters.

        Add to the existing filter, perhaps to exclude an additional directory.

        Args:
            endswith: Passed through to FileFilter.
            exclude: Passed through to FileFilter.

        Returns a new check.
        """
        return self.with_file_filter(
            FileFilter(endswith=_make_str_tuple(endswith), exclude=exclude)
        )

    def with_file_filter(self, file_filter: FileFilter) -> Check:
        """Create a new check identical to this one, but with extra filters.

        Add to the existing filter, perhaps to exclude an additional directory.

        Args:
            file_filter: Additional filter rules.

        Returns a new check.
        """
        clone = copy.copy(self)
        if clone.filter:
            clone.filter.exclude = clone.filter.exclude + file_filter.exclude
            clone.filter.endswith = clone.filter.endswith + file_filter.endswith
            clone.filter.name = file_filter.name or clone.filter.name
            clone.filter.suffix = clone.filter.suffix + file_filter.suffix
        else:
            clone.filter = file_filter
        return clone

    def run(
        self,
        ctx: PresubmitContext,
        count: int,
        total: int,
        substep: Optional[str] = None,
    ) -> PresubmitResult:
        """Runs the presubmit check on the provided paths."""

        _print_ui(
            _box(
                _CHECK_UPPER,
                f'{count}/{total}',
                self.name,
                plural(ctx.paths, "file"),
            )
        )

        substep_part = f'.{substep}' if substep else ''
        _LOG.debug(
            '[%d/%d] Running %s%s on %s',
            count,
            total,
            self.name,
            substep_part,
            plural(ctx.paths, "file"),
        )

        start_time_s = time.time()
        result: PresubmitResult
        if substep:
            result = self.run_substep(ctx, substep)
        else:
            result = self(ctx)
        time_str = _format_time(time.time() - start_time_s)
        _LOG.debug('%s %s', self.name, result.value)

        _print_ui(
            _box(_CHECK_LOWER, result.colorized(_LEFT), self.name, time_str)
        )
        _LOG.debug('%s duration:%s', self.name, time_str)

        return result

    def _try_call(
        self,
        func: Callable,
        ctx,
        *args,
        **kwargs,
    ) -> PresubmitResult:
        try:
            result = func(ctx, *args, **kwargs)
            if ctx.failed:
                return PresubmitResult.FAIL
            if isinstance(result, PresubmitResult):
                return result
            return PresubmitResult.PASS

        except PresubmitFailure as failure:
            if str(failure):
                _LOG.warning('%s', failure)
            return PresubmitResult.FAIL

        except Exception as _failure:  # pylint: disable=broad-except
            _LOG.exception('Presubmit check %s failed!', self.name)
            return PresubmitResult.FAIL

        except KeyboardInterrupt:
            _print_ui()
            return PresubmitResult.CANCEL

    def run_substep(
        self, ctx: PresubmitContext, name: Optional[str]
    ) -> PresubmitResult:
        for substep in self.substeps():
            if substep.name == name:
                return substep(ctx)

        expected = ', '.join(repr(s.name) for s in self.substeps())
        raise LookupError(f'bad substep name: {name!r} (expected: {expected})')

    def __call__(self, ctx: PresubmitContext) -> PresubmitResult:
        """Calling a Check calls its underlying substeps directly.

        This makes it possible to call functions wrapped by @filter_paths. The
        prior filters are ignored, so new filters may be applied.
        """
        result: PresubmitResult
        for substep in self.substeps():
            result = self._try_call(substep, ctx)
            if result and result != PresubmitResult.PASS:
                return result
        return PresubmitResult.PASS


def _required_args(function: Callable) -> Iterable[Parameter]:
    """Returns the required arguments for a function."""
    optional_types = Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD

    for param in signature(function).parameters.values():
        if param.default is param.empty and param.kind not in optional_types:
            yield param


def _ensure_is_valid_presubmit_check_function(chk: Callable) -> None:
    """Checks if a Callable can be used as a presubmit check."""
    try:
        required_args = tuple(_required_args(chk))
    except (TypeError, ValueError):
        raise TypeError(
            'Presubmit checks must be callable, but '
            f'{chk!r} is a {type(chk).__name__}'
        )

    if len(required_args) != 1:
        raise TypeError(
            f'Presubmit check functions must have exactly one required '
            f'positional argument (the PresubmitContext), but '
            f'{chk.__name__} has {len(required_args)} required arguments'
            + (
                f' ({", ".join(a.name for a in required_args)})'
                if required_args
                else ''
            )
        )


def filter_paths(
    *,
    endswith: Iterable[str] = (),
    exclude: Iterable[Union[Pattern[str], str]] = (),
    file_filter: Optional[FileFilter] = None,
    always_run: bool = False,
) -> Callable[[Callable], Check]:
    """Decorator for filtering the paths list for a presubmit check function.

    Path filters only apply when the function is used as a presubmit check.
    Filters are ignored when the functions are called directly. This makes it
    possible to reuse functions wrapped in @filter_paths in other presubmit
    checks, potentially with different path filtering rules.

    Args:
        endswith: str or iterable of path endings to include
        exclude: regular expressions of paths to exclude
        file_filter: FileFilter used to select files
        always_run: Run check even when no files match
    Returns:
        a wrapped version of the presubmit function
    """

    if file_filter:
        real_file_filter = file_filter
        if endswith or exclude:
            raise ValueError(
                'Must specify either file_filter or '
                'endswith/exclude args, not both'
            )
    else:
        # TODO(b/238426363): Remove these arguments and use FileFilter only.
        real_file_filter = FileFilter(
            endswith=_make_str_tuple(endswith), exclude=exclude
        )

    def filter_paths_for_function(function: Callable):
        return Check(function, real_file_filter, always_run=always_run)

    return filter_paths_for_function


def call(*args, **kwargs) -> None:
    """Optional subprocess wrapper that causes a PresubmitFailure on errors."""
    attributes, command = tools.format_command(args, kwargs)
    _LOG.debug('[RUN] %s\n%s', attributes, command)

    tee = kwargs.pop('tee', None)
    propagate_sigterm = kwargs.pop('propagate_sigterm', False)

    env = pw_cli.env.pigweed_environment()
    kwargs['stdout'] = subprocess.PIPE
    kwargs['stderr'] = subprocess.STDOUT

    process = subprocess.Popen(args, **kwargs)
    assert process.stdout

    # Set up signal handler if requested.
    signaled = False
    if propagate_sigterm:

        def signal_handler(_signal_number: int, _stack_frame: Any) -> None:
            nonlocal signaled
            signaled = True
            process.terminate()

        previous_signal_handler = signal.signal(signal.SIGTERM, signal_handler)

    if env.PW_PRESUBMIT_DISABLE_SUBPROCESS_CAPTURE:
        while True:
            line = process.stdout.readline().decode(errors='backslashreplace')
            if not line:
                break
            _LOG.info(line.rstrip())
            if tee:
                tee.write(line)

    stdout, _ = process.communicate()
    if tee:
        tee.write(stdout.decode(errors='backslashreplace'))

    logfunc = _LOG.warning if process.returncode else _LOG.debug
    logfunc('[FINISHED]\n%s', command)
    logfunc(
        '[RESULT] %s with return code %d',
        'Failed' if process.returncode else 'Passed',
        process.returncode,
    )
    if stdout:
        logfunc('[OUTPUT]\n%s', stdout.decode(errors='backslashreplace'))

    if propagate_sigterm:
        signal.signal(signal.SIGTERM, previous_signal_handler)
        if signaled:
            _LOG.warning('Exiting due to SIGTERM.')
            sys.exit(1)

    if process.returncode:
        raise PresubmitFailure


def install_package(
    ctx: Union[FormatContext, PresubmitContext],
    name: str,
    force: bool = False,
) -> None:
    """Install package with given name in given path."""
    root = ctx.package_root
    mgr = package_manager.PackageManager(root)

    if not mgr.list():
        raise PresubmitFailure(
            'no packages configured, please import your pw_package '
            'configuration module'
        )

    if not mgr.status(name) or force:
        mgr.install(name, force=force)



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Dialog box for log selection functions."""

from __future__ import annotations
import functools
from typing import TYPE_CHECKING

from prompt_toolkit.filters import Condition
from prompt_toolkit.layout import (
    ConditionalContainer,
    FormattedTextControl,
    Window,
    WindowAlign,
)

from pw_console.widgets import (
    create_border,
    mouse_handlers,
    to_checkbox_with_keybind_indicator,
    to_keybind_indicator,
)

if TYPE_CHECKING:
    from pw_console.log_pane import LogPane


class LogPaneSelectionDialog(ConditionalContainer):
    """Dialog box for showing log selection functions.

    Displays number of lines selected, buttons for copying to the clipboar or
    saving to a file, and buttons to select all or cancel (clear) the
    selection."""

    # Height of the dialog box contens in lines of text.
    DIALOG_HEIGHT = 3

    def __init__(self, log_pane: 'LogPane'):
        self.log_pane = log_pane
        self.log_view = log_pane.log_view

        self._markdown_flag: bool = False
        self._table_flag: bool = True

        selection_bar_control = FormattedTextControl(self.get_fragments)
        selection_bar_window = Window(
            content=selection_bar_control,
            height=1,
            align=WindowAlign.LEFT,
            dont_extend_width=False,
            style='class:selection-dialog',
        )

        super().__init__(
            create_border(
                selection_bar_window,
                (LogPaneSelectionDialog.DIALOG_HEIGHT - 1),
                border_style='class:selection-dialog-border',
                base_style='class:selection-dialog-default-fg',
                top=False,
                right=False,
            ),
            filter=Condition(lambda: self.log_view.visual_select_mode),
        )

    def focus_log_pane(self):
        self.log_pane.application.focus_on_container(self.log_pane)

    def _toggle_markdown_flag(self) -> None:
        self._markdown_flag = not self._markdown_flag

    def _toggle_table_flag(self) -> None:
        self._table_flag = not self._table_flag

    def _select_all(self) -> None:
        self.log_view.visual_select_all()

    def _select_none(self) -> None:
        self.log_view.clear_visual_selection()

    def _copy_selection(self) -> None:
        if self.log_view.export_logs(
            to_clipboard=True,
            use_table_formatting=self._table_flag,
            selected_lines_only=True,
            add_markdown_fence=self._markdown_flag,
        ):
            self._select_none()

    def _saveas_file(self) -> None:
        self.log_pane.start_saveas(
            table_format=self._table_flag, selected_lines_only=True
        )

    def get_fragments(self):
        """Return formatted text tuples for both rows of the selection
        dialog."""

        focus = functools.partial(mouse_handlers.on_click, self.focus_log_pane)

        one_space = ('', ' ', focus)
        two_spaces = ('', '  ', focus)
        select_all = functools.partial(
            mouse_handlers.on_click, self._select_all
        )
        select_none = functools.partial(
            mouse_handlers.on_click, self._select_none
        )

        copy_selection = functools.partial(
            mouse_handlers.on_click, self._copy_selection
        )
        saveas_file = functools.partial(
            mouse_handlers.on_click, self._saveas_file
        )
        toggle_markdown = functools.partial(
            mouse_handlers.on_click,
            self._toggle_markdown_flag,
        )
        toggle_table = functools.partial(
            mouse_handlers.on_click, self._toggle_table_flag
        )

        button_style = 'class:toolbar-button-inactive'

        # First row of text
        fragments = [
            (
                'class:selection-dialog-title',
                ' {} Selected '.format(
                    self.log_view.visual_selected_log_count()
                ),
                focus,
            ),
            one_space,
            ('class:selection-dialog-default-fg', 'Format: ', focus),
        ]

        # Table and Markdown options
        fragments.extend(
            to_checkbox_with_keybind_indicator(
                self._table_flag,
                key='',
                description='Table',
                mouse_handler=toggle_table,
                base_style='class:selection-dialog-default-bg',
            )
        )

        fragments.extend(
            to_checkbox_with_keybind_indicator(
                self._markdown_flag,
                key='',
                description='Markdown',
                mouse_handler=toggle_markdown,
                base_style='class:selection-dialog-default-bg',
            )
        )

        # Line break
        fragments.append(('', '\n'))

        # Second row of text
        fragments.append(one_space)

        fragments.extend(
            to_keybind_indicator(
                key='Ctrl-c',
                description='Cancel',
                mouse_handler=select_none,
                base_style=button_style,
            )
        )
        fragments.append(two_spaces)

        fragments.extend(
            to_keybind_indicator(
                key='Ctrl-a',
                description='Select All',
                mouse_handler=select_all,
                base_style=button_style,
            )
        )
        fragments.append(two_spaces)

        fragments.append(one_space)
        fragments.extend(
            to_keybind_indicator(
                key='',
                description='Save as File',
                mouse_handler=saveas_file,
                base_style=button_style,
            )
        )
        fragments.append(two_spaces)

        fragments.extend(
            to_keybind_indicator(
                key='',
                description='Copy',
                mouse_handler=copy_selection,
                base_style=button_style,
            )
        )
        fragments.append(one_space)

        return fragments



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Finds components for a given manifest."""

from typing import Any, List, Optional, Tuple

import pathlib
import sys
import xml.etree.ElementTree


def _gn_str_out(name: str, val: Any):
    """Outputs scoped string in GN format."""
    print(f'{name} = "{val}"')


def _gn_list_str_out(name: str, val: List[Any]):
    """Outputs list of strings in GN format with correct escaping."""
    list_str = ','.join(
        '"' + str(x).replace('"', r'\"').replace('$', r'\$') + '"' for x in val
    )
    print(f'{name} = [{list_str}]')


def _gn_list_path_out(
    name: str, val: List[pathlib.Path], path_prefix: Optional[str] = None
):
    """Outputs list of paths in GN format with common prefix."""
    if path_prefix is not None:
        str_val = list(f'{path_prefix}/{str(d)}' for d in val)
    else:
        str_val = list(str(d) for d in val)
    _gn_list_str_out(name, str_val)


def get_component(
    root: xml.etree.ElementTree.Element, component_id: str
) -> Tuple[Optional[xml.etree.ElementTree.Element], Optional[pathlib.Path]]:
    """Parse <component> manifest stanza.

    Schema:
        <component id="{component_id}" package_base_path="component">
        </component>

    Args:
        root: root of element tree.
        component_id: id of component to return.

    Returns:
        (element, base_path) for the component, or (None, None).
    """
    xpath = f'./components/component[@id="{component_id}"]'
    component = root.find(xpath)
    if component is None:
        return (None, None)

    try:
        base_path = pathlib.Path(component.attrib['package_base_path'])
        return (component, base_path)
    except KeyError:
        return (component, None)


def parse_defines(
    root: xml.etree.ElementTree.Element, component_id: str
) -> List[str]:
    """Parse pre-processor definitions for a component.

    Schema:
        <defines>
          <define name="EXAMPLE" value="1"/>
          <define name="OTHER"/>
        </defines>

    Args:
        root: root of element tree.
        component_id: id of component to return.

    Returns:
        list of str NAME=VALUE or NAME for the component.
    """
    xpath = f'./components/component[@id="{component_id}"]/defines/define'
    return list(_parse_define(define) for define in root.findall(xpath))


def _parse_define(define: xml.etree.ElementTree.Element) -> str:
    """Parse <define> manifest stanza.

    Schema:
        <define name="EXAMPLE" value="1"/>
        <define name="OTHER"/>

    Args:
        define: XML Element for <define>.

    Returns:
        str with a value NAME=VALUE or NAME.
    """
    name = define.attrib['name']
    value = define.attrib.get('value', None)
    if value is None:
        return name

    return f'{name}={value}'


def parse_include_paths(
    root: xml.etree.ElementTree.Element, component_id: str
) -> List[pathlib.Path]:
    """Parse include directories for a component.

    Schema:
        <component id="{component_id}" package_base_path="component">
          <include_paths>
            <include_path relative_path="./" type="c_include"/>
          </include_paths>
        </component>

    Args:
        root: root of element tree.
        component_id: id of component to return.

    Returns:
        list of include directories for the component.
    """
    (component, base_path) = get_component(root, component_id)
    if component is None:
        return []

    include_paths: List[pathlib.Path] = []
    for include_type in ('c_include', 'asm_include'):
        include_xpath = f'./include_paths/include_path[@type="{include_type}"]'

        include_paths.extend(
            _parse_include_path(include_path, base_path)
            for include_path in component.findall(include_xpath)
        )
    return include_paths


def _parse_include_path(
    include_path: xml.etree.ElementTree.Element,
    base_path: Optional[pathlib.Path],
) -> pathlib.Path:
    """Parse <include_path> manifest stanza.

    Schema:
        <include_path relative_path="./" type="c_include"/>

    Args:
        include_path: XML Element for <input_path>.
        base_path: prefix for paths.

    Returns:
        Path, prefixed with `base_path`.
    """
    path = pathlib.Path(include_path.attrib['relative_path'])
    if base_path is None:
        return path
    return base_path / path


def parse_headers(
    root: xml.etree.ElementTree.Element, component_id: str
) -> List[pathlib.Path]:
    """Parse header files for a component.

    Schema:
        <component id="{component_id}" package_base_path="component">
          <source relative_path="./" type="c_include">
            <files mask="example.h"/>
          </source>
        </component>

    Args:
        root: root of element tree.
        component_id: id of component to return.

    Returns:
        list of header files for the component.
    """
    return _parse_sources(root, component_id, 'c_include')


def parse_sources(
    root: xml.etree.ElementTree.Element, component_id: str
) -> List[pathlib.Path]:
    """Parse source files for a component.

    Schema:
        <component id="{component_id}" package_base_path="component">
          <source relative_path="./" type="src">
            <files mask="example.cc"/>
          </source>
        </component>

    Args:
        root: root of element tree.
        component_id: id of component to return.

    Returns:
        list of source files for the component.
    """
    source_files = []
    for source_type in ('src', 'src_c', 'src_cpp', 'asm_include'):
        source_files.extend(_parse_sources(root, component_id, source_type))
    return source_files


def parse_libs(
    root: xml.etree.ElementTree.Element, component_id: str
) -> List[pathlib.Path]:
    """Parse pre-compiled libraries for a component.

    Schema:
        <component id="{component_id}" package_base_path="component">
          <source relative_path="./" type="lib">
            <files mask="example.a"/>
          </source>
        </component>

    Args:
        root: root of element tree.
        component_id: id of component to return.

    Returns:
        list of pre-compiler libraries for the component.
    """
    return _parse_sources(root, component_id, 'lib')


def _parse_sources(
    root: xml.etree.ElementTree.Element, component_id: str, source_type: str
) -> List[pathlib.Path]:
    """Parse <source> manifest stanza.

    Schema:
        <component id="{component_id}" package_base_path="component">
          <source relative_path="./" type="{source_type}">
            <files mask="example.h"/>
          </source>
        </component>

    Args:
        root: root of element tree.
        component_id: id of component to return.
        source_type: type of source to search for.

    Returns:
        list of source files for the component.
    """
    (component, base_path) = get_component(root, component_id)
    if component is None:
        return []

    sources: List[pathlib.Path] = []
    source_xpath = f'./source[@type="{source_type}"]'
    for source in component.findall(source_xpath):
        relative_path = pathlib.Path(source.attrib['relative_path'])
        if base_path is not None:
            relative_path = base_path / relative_path

        sources.extend(
            relative_path / files.attrib['mask']
            for files in source.findall('./files')
        )
    return sources


def parse_dependencies(
    root: xml.etree.ElementTree.Element, component_id: str
) -> List[str]:
    """Parse the list of dependencies for a component.

    Optional dependencies are ignored for parsing since they have to be
    included explicitly.

    Schema:
        <dependencies>
          <all>
            <component_dependency value="component"/>
            <component_dependency value="component"/>
            <any_of>
              <component_dependency value="component"/>
              <component_dependency value="component"/>
            </any_of>
          </all>
        </dependencies>

    Args:
        root: root of element tree.
        component_id: id of component to return.

    Returns:
        list of component id dependencies of the component.
    """
    dependencies = []
    xpath = f'./components/component[@id="{component_id}"]/dependencies/*'
    for dependency in root.findall(xpath):
        dependencies.extend(_parse_dependency(dependency))
    return dependencies


def _parse_dependency(dependency: xml.etree.ElementTree.Element) -> List[str]:
    """Parse <all>, <any_of>, and <component_dependency> manifest stanzas.

    Schema:
        <all>
          <component_dependency value="component"/>
          <component_dependency value="component"/>
          <any_of>
            <component_dependency value="component"/>
            <component_dependency value="component"/>
          </any_of>
        </all>

    Args:
        dependency: XML Element of dependency.

    Returns:
        list of component id dependencies.
    """
    if dependency.tag == 'component_dependency':
        return [dependency.attrib['value']]
    if dependency.tag == 'all':
        dependencies = []
        for subdependency in dependency:
            dependencies.extend(_parse_dependency(subdependency))
        return dependencies
    if dependency.tag == 'any_of':
        # Explicitly ignore.
        return []

    # Unknown dependency tag type.
    return []


def check_dependencies(
    root: xml.etree.ElementTree.Element,
    component_id: str,
    include: List[str],
    exclude: Optional[List[str]] = None,
) -> bool:
    """Check the list of optional dependencies for a component.

    Verifies that the optional dependencies for a component are satisfied by
    components listed in `include` or `exclude`.

    Args:
        root: root of element tree.
        component_id: id of component to check.
        include: list of component ids included in the project.
        exclude: list of component ids explicitly excluded from the project.

    Returns:
        True if dependencies are satisfied, False if not.
    """
    xpath = f'./components/component[@id="{component_id}"]/dependencies/*'
    for dependency in root.findall(xpath):
        if not _check_dependency(dependency, include, exclude=exclude):
            return False
    return True


def _check_dependency(
    dependency: xml.etree.ElementTree.Element,
    include: List[str],
    exclude: Optional[List[str]] = None,
) -> bool:
    """Check a dependency for a component.

    Verifies that the given {dependency} is satisfied by components listed in
    `include` or `exclude`.

    Args:
        dependency: XML Element of dependency.
        include: list of component ids included in the project.
        exclude: list of component ids explicitly excluded from the project.

    Returns:
        True if dependencies are satisfied, False if not.
    """
    if dependency.tag == 'component_dependency':
        component_id = dependency.attrib['value']
        return component_id in include or (
            exclude is not None and component_id in exclude
        )
    if dependency.tag == 'all':
        for subdependency in dependency:
            if not _check_dependency(subdependency, include, exclude=exclude):
                return False
        return True
    if dependency.tag == 'any_of':
        for subdependency in dependency:
            if _check_dependency(subdependency, include, exclude=exclude):
                return True

        tree = xml.etree.ElementTree.tostring(dependency).decode('utf-8')
        print(f'Unsatisfied dependency from: {tree}', file=sys.stderr)
        return False

    # Unknown dependency tag type.
    return True


def create_project(
    root: xml.etree.ElementTree.Element,
    include: List[str],
    exclude: Optional[List[str]] = None,
) -> Tuple[
    List[str],
    List[str],
    List[pathlib.Path],
    List[pathlib.Path],
    List[pathlib.Path],
    List[pathlib.Path],
]:
    """Create a project from a list of specified components.

    Args:
        root: root of element tree.
        include: list of component ids included in the project.
        exclude: list of component ids excluded from the project.

    Returns:
        (component_ids, defines, include_paths, headers, sources, libs) for the
        project.
    """
    # Build the project list from the list of included components by expanding
    # dependencies.
    project_list = []
    pending_list = include
    while len(pending_list) > 0:
        component_id = pending_list.pop(0)
        if component_id in project_list:
            continue
        if exclude is not None and component_id in exclude:
            continue

        project_list.append(component_id)
        pending_list.extend(parse_dependencies(root, component_id))

    return (
        project_list,
        sum(
            (
                parse_defines(root, component_id)
                for component_id in project_list
            ),
            [],
        ),
        sum(
            (
                parse_include_paths(root, component_id)
                for component_id in project_list
            ),
            [],
        ),
        sum(
            (
                parse_headers(root, component_id)
                for component_id in project_list
            ),
            [],
        ),
        sum(
            (
                parse_sources(root, component_id)
                for component_id in project_list
            ),
            [],
        ),
        sum(
            (parse_libs(root, component_id) for component_id in project_list),
            [],
        ),
    )


def project(
    manifest_path: pathlib.Path,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    path_prefix: Optional[str] = None,
):
    """Output GN scope for a project with the specified components.

    Args:
        manifest_path: path to SDK manifest XML.
        include: list of component ids included in the project.
        exclude: list of component ids excluded from the project.
        path_prefix: string prefix to prepend to all paths.
    """
    assert include is not None, "Project must include at least one component."

    tree = xml.etree.ElementTree.parse(manifest_path)
    root = tree.getroot()

    (
        component_ids,
        defines,
        include_dirs,
        headers,
        sources,
        libs,
    ) = create_project(root, include, exclude=exclude)

    for component_id in component_ids:
        if not check_dependencies(
            root, component_id, component_ids, exclude=exclude
        ):
            return

    _gn_list_str_out('defines', defines)
    _gn_list_path_out('include_dirs', include_dirs, path_prefix=path_prefix)
    _gn_list_path_out('public', headers, path_prefix=path_prefix)
    _gn_list_path_out('sources', sources, path_prefix=path_prefix)
    _gn_list_path_out('libs', libs, path_prefix=path_prefix)



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""LogPane Save As Dialog."""

from __future__ import annotations
import functools
from pathlib import Path
from typing import Optional, TYPE_CHECKING

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from prompt_toolkit.completion import PathCompleter
from prompt_toolkit.filters import Condition
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.layout import (
    ConditionalContainer,
    FormattedTextControl,
    HSplit,
    Window,
    WindowAlign,
)
from prompt_toolkit.widgets import TextArea
from prompt_toolkit.validation import (
    ValidationError,
    Validator,
)

from pw_console.widgets import (
    create_border,
    mouse_handlers,
    to_checkbox_with_keybind_indicator,
    to_keybind_indicator,
)

if TYPE_CHECKING:
    from pw_console.log_pane import LogPane


class PathValidator(Validator):
    """Validation of file path input."""

    def validate(self, document):
        """Check input path leads to a valid parent directory."""
        target_path = Path(document.text).expanduser()

        if not target_path.parent.exists():
            raise ValidationError(
                # Set cursor position to the end
                len(document.text),
                "Directory doesn't exist: %s" % document.text,
            )

        if target_path.is_dir():
            raise ValidationError(
                # Set cursor position to the end
                len(document.text),
                "File input is an existing directory: %s" % document.text,
            )


class LogPaneSaveAsDialog(ConditionalContainer):
    """Dialog box for saving logs to a file."""

    # Height of the dialog box contens in lines of text.
    DIALOG_HEIGHT = 3

    def __init__(self, log_pane: 'LogPane'):
        self.log_pane = log_pane

        self.path_validator = PathValidator()

        self._export_with_table_formatting: bool = True
        self._export_with_selected_lines_only: bool = False

        self.starting_file_path: str = str(Path.cwd())

        self.input_field = TextArea(
            prompt=[
                (
                    'class:saveas-dialog-setting',
                    'File: ',
                    functools.partial(
                        mouse_handlers.on_click,
                        self.focus_self,
                    ),
                )
            ],
            # Pre-fill the current working directory.
            text=self.starting_file_path,
            focusable=True,
            focus_on_click=True,
            scrollbar=False,
            multiline=False,
            height=1,
            dont_extend_height=True,
            dont_extend_width=False,
            accept_handler=self._saveas_accept_handler,
            validator=self.path_validator,
            history=InMemoryHistory(),
            completer=PathCompleter(expanduser=True),
        )

        self.input_field.buffer.cursor_position = len(self.starting_file_path)

        settings_bar_control = FormattedTextControl(self.get_settings_fragments)
        settings_bar_window = Window(
            content=settings_bar_control,
            height=1,
            align=WindowAlign.LEFT,
            dont_extend_width=False,
        )

        action_bar_control = FormattedTextControl(self.get_action_fragments)
        action_bar_window = Window(
            content=action_bar_control,
            height=1,
            align=WindowAlign.RIGHT,
            dont_extend_width=False,
        )

        # Add additional keybindings for the input_field text area.
        key_bindings = KeyBindings()
        register = self.log_pane.application.prefs.register_keybinding

        @register('save-as-dialog.cancel', key_bindings)
        def _close_saveas_dialog(_event: KeyPressEvent) -> None:
            """Close save as dialog."""
            self.close_dialog()

        self.input_field.control.key_bindings = key_bindings

        super().__init__(
            create_border(
                HSplit(
                    [
                        settings_bar_window,
                        self.input_field,
                        action_bar_window,
                    ],
                    height=LogPaneSaveAsDialog.DIALOG_HEIGHT,
                    style='class:saveas-dialog',
                ),
                LogPaneSaveAsDialog.DIALOG_HEIGHT,
                border_style='class:saveas-dialog-border',
                left_margin_columns=1,
            ),
            filter=Condition(lambda: self.log_pane.saveas_dialog_active),
        )

    def focus_self(self):
        self.log_pane.application.application.layout.focus(self)

    def close_dialog(self):
        """Close this dialog."""
        self.log_pane.saveas_dialog_active = False
        self.log_pane.application.focus_on_container(self.log_pane)
        self.log_pane.redraw_ui()

    def _toggle_table_formatting(self):
        self._export_with_table_formatting = (
            not self._export_with_table_formatting
        )

    def _toggle_selected_lines(self):
        self._export_with_selected_lines_only = (
            not self._export_with_selected_lines_only
        )

    def set_export_options(
        self,
        table_format: Optional[bool] = None,
        selected_lines_only: Optional[bool] = None,
    ) -> None:
        # Allows external callers such as the line selection dialog to set
        # export format options.
        if table_format is not None:
            self._export_with_table_formatting = table_format

        if selected_lines_only is not None:
            self._export_with_selected_lines_only = selected_lines_only

    def save_action(self):
        """Trigger save file execution on mouse click.

        This ultimately runs LogPaneSaveAsDialog._saveas_accept_handler()."""
        self.input_field.buffer.validate_and_handle()

    def _saveas_accept_handler(self, buff: Buffer) -> bool:
        """Function run when hitting Enter in the input_field."""
        input_text = buff.text
        if len(input_text) == 0:
            self.close_dialog()
            # Don't save anything if empty input.
            return False

        if self.log_pane.log_view.export_logs(
            file_name=input_text,
            use_table_formatting=self._export_with_table_formatting,
            selected_lines_only=self._export_with_selected_lines_only,
        ):
            self.close_dialog()
            # Reset selected_lines_only
            self.set_export_options(selected_lines_only=False)
            # Erase existing input text.
            return False

        # Keep existing text if error
        return True

    def get_settings_fragments(self):
        """Return FormattedText with current save settings."""
        # Mouse handlers
        focus = functools.partial(mouse_handlers.on_click, self.focus_self)
        toggle_table_formatting = functools.partial(
            mouse_handlers.on_click,
            self._toggle_table_formatting,
        )
        toggle_selected_lines = functools.partial(
            mouse_handlers.on_click,
            self._toggle_selected_lines,
        )

        # Separator should have the focus mouse handler so clicking on any
        # whitespace focuses the input field.
        separator_text = ('', '  ', focus)

        # Default button style
        button_style = 'class:toolbar-button-inactive'

        fragments = [('class:saveas-dialog-title', 'Save as File', focus)]
        fragments.append(separator_text)

        # Table checkbox
        fragments.extend(
            to_checkbox_with_keybind_indicator(
                checked=self._export_with_table_formatting,
                key='',  # No key shortcut help text
                description='Table Formatting',
                mouse_handler=toggle_table_formatting,
                base_style=button_style,
            )
        )

        # Two space separator
        fragments.append(separator_text)

        # Selected lines checkbox
        fragments.extend(
            to_checkbox_with_keybind_indicator(
                checked=self._export_with_selected_lines_only,
                key='',  # No key shortcut help text
                description='Selected Lines Only',
                mouse_handler=toggle_selected_lines,
                base_style=button_style,
            )
        )

        # Two space separator
        fragments.append(separator_text)

        return fragments

    def get_action_fragments(self):
        """Return FormattedText with the save action buttons."""
        # Mouse handlers
        focus = functools.partial(mouse_handlers.on_click, self.focus_self)
        cancel = functools.partial(mouse_handlers.on_click, self.close_dialog)
        save = functools.partial(mouse_handlers.on_click, self.save_action)

        # Separator should have the focus mouse handler so clicking on any
        # whitespace focuses the input field.
        separator_text = ('', '  ', focus)

        # Default button style
        button_style = 'class:toolbar-button-inactive'

        fragments = [separator_text]
        # Cancel button
        fragments.extend(
            to_keybind_indicator(
                key='Ctrl-c',
                description='Cancel',
                mouse_handler=cancel,
                base_style=button_style,
            )
        )

        # Two space separator
        fragments.append(separator_text)

        # Save button
        fragments.extend(
            to_keybind_indicator(
                key='Enter',
                description='Save',
                mouse_handler=save,
                base_style=button_style,
            )
        )

        # One space separator
        fragments.append(('', ' ', focus))

        return fragments



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_software_update/bundled_update.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_protobuf_protos import common_pb2 as pw__protobuf__protos_dot_common__pb2
from pw_tokenizer.proto import options_pb2 as pw__tokenizer_dot_proto_dot_options__pb2
from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_software_update/bundled_update.proto',
  package='pw.software_update',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\'pw_software_update/bundled_update.proto\x12\x12pw.software_update\x1a\x1fpw_protobuf_protos/common.proto\x1a pw_tokenizer/proto/options.proto\x1a\x19google/protobuf/any.proto\"\xa2\x01\n\x12\x42undledUpdateState\"\x8b\x01\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0c\n\x08INACTIVE\x10\x01\x12\x10\n\x0cTRANSFERRING\x10\x02\x12\x0f\n\x0bTRANSFERRED\x10\x03\x12\r\n\tVERIFYING\x10\x04\x12\x0c\n\x08VERIFIED\x10\x05\x12\x0c\n\x08\x41PPLYING\x10\x06\x12\x0c\n\x08\x41\x42ORTING\x10\x07\x12\x0c\n\x08\x46INISHED\x10\x08\"\x91\x01\n\x13\x42undledUpdateResult\"z\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07SUCCESS\x10\x01\x12\x11\n\rUNKNOWN_ERROR\x10\x02\x12\x0b\n\x07\x41\x42ORTED\x10\x03\x12\x13\n\x0fTRANSFER_FAILED\x10\x04\x12\x11\n\rVERIFY_FAILED\x10\x05\x12\x10\n\x0c\x41PPLY_FAILED\x10\x06\"\xc7\x03\n\x13\x42undledUpdateStatus\x12:\n\x05state\x18\x01 \x01(\x0e\x32+.pw.software_update.BundledUpdateState.Enum\x12\x41\n\x06result\x18\x02 \x01(\x0e\x32,.pw.software_update.BundledUpdateResult.EnumH\x00\x88\x01\x01\x12\x34\n\'current_state_progress_hundreth_percent\x18\x03 \x01(\rH\x01\x88\x01\x01\x12\x18\n\x0btransfer_id\x18\x04 \x01(\rH\x02\x88\x01\x01\x12\x1c\n\x0f\x62undle_filename\x18\x05 \x01(\tH\x03\x88\x01\x01\x12\x17\n\x04note\x18\x06 \x01(\x0c\x42\x04\x80\xaf&\x00H\x04\x88\x01\x01\x12\x32\n\x0f\x65xtended_status\x18\x07 \x01(\x0b\x32\x14.google.protobuf.AnyH\x05\x88\x01\x01\x42\t\n\x07_resultB*\n(_current_state_progress_hundreth_percentB\x0e\n\x0c_transfer_idB\x12\n\x10_bundle_filenameB\x07\n\x05_noteB\x12\n\x10_extended_status\"@\n\x0cStartRequest\x12\x1c\n\x0f\x62undle_filename\x18\x01 \x01(\tH\x00\x88\x01\x01\x42\x12\n\x10_bundle_filename2\x95\x04\n\rBundledUpdate\x12H\n\tGetStatus\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12R\n\x05Start\x12 .pw.software_update.StartRequest\x1a\'.pw.software_update.BundledUpdateStatus\x12M\n\x0eSetTransferred\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x45\n\x06Verify\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x44\n\x05\x41pply\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x44\n\x05\x41\x62ort\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x44\n\x05Reset\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatusb\x06proto3'
  ,
  dependencies=[pw__protobuf__protos_dot_common__pb2.DESCRIPTOR,pw__tokenizer_dot_proto_dot_options__pb2.DESCRIPTOR,google_dot_protobuf_dot_any__pb2.DESCRIPTOR,])



_BUNDLEDUPDATESTATE_ENUM = _descriptor.EnumDescriptor(
  name='Enum',
  full_name='pw.software_update.BundledUpdateState.Enum',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INACTIVE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRANSFERRING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRANSFERRED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='VERIFYING', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='VERIFIED', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='APPLYING', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ABORTING', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FINISHED', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=181,
  serialized_end=320,
)
_sym_db.RegisterEnumDescriptor(_BUNDLEDUPDATESTATE_ENUM)

_BUNDLEDUPDATERESULT_ENUM = _descriptor.EnumDescriptor(
  name='Enum',
  full_name='pw.software_update.BundledUpdateResult.Enum',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUCCESS', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_ERROR', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ABORTED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRANSFER_FAILED', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='VERIFY_FAILED', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='APPLY_FAILED', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=346,
  serialized_end=468,
)
_sym_db.RegisterEnumDescriptor(_BUNDLEDUPDATERESULT_ENUM)


_BUNDLEDUPDATESTATE = _descriptor.Descriptor(
  name='BundledUpdateState',
  full_name='pw.software_update.BundledUpdateState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _BUNDLEDUPDATESTATE_ENUM,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=158,
  serialized_end=320,
)


_BUNDLEDUPDATERESULT = _descriptor.Descriptor(
  name='BundledUpdateResult',
  full_name='pw.software_update.BundledUpdateResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _BUNDLEDUPDATERESULT_ENUM,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=323,
  serialized_end=468,
)


_BUNDLEDUPDATESTATUS = _descriptor.Descriptor(
  name='BundledUpdateStatus',
  full_name='pw.software_update.BundledUpdateStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='state', full_name='pw.software_update.BundledUpdateStatus.state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='result', full_name='pw.software_update.BundledUpdateStatus.result', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='current_state_progress_hundreth_percent', full_name='pw.software_update.BundledUpdateStatus.current_state_progress_hundreth_percent', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='transfer_id', full_name='pw.software_update.BundledUpdateStatus.transfer_id', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bundle_filename', full_name='pw.software_update.BundledUpdateStatus.bundle_filename', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='note', full_name='pw.software_update.BundledUpdateStatus.note', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='extended_status', full_name='pw.software_update.BundledUpdateStatus.extended_status', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_result', full_name='pw.software_update.BundledUpdateStatus._result',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_current_state_progress_hundreth_percent', full_name='pw.software_update.BundledUpdateStatus._current_state_progress_hundreth_percent',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_transfer_id', full_name='pw.software_update.BundledUpdateStatus._transfer_id',
      index=2, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_bundle_filename', full_name='pw.software_update.BundledUpdateStatus._bundle_filename',
      index=3, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_note', full_name='pw.software_update.BundledUpdateStatus._note',
      index=4, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_extended_status', full_name='pw.software_update.BundledUpdateStatus._extended_status',
      index=5, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=471,
  serialized_end=926,
)


_STARTREQUEST = _descriptor.Descriptor(
  name='StartRequest',
  full_name='pw.software_update.StartRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='bundle_filename', full_name='pw.software_update.StartRequest.bundle_filename', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_bundle_filename', full_name='pw.software_update.StartRequest._bundle_filename',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=928,
  serialized_end=992,
)

_BUNDLEDUPDATESTATE_ENUM.containing_type = _BUNDLEDUPDATESTATE
_BUNDLEDUPDATERESULT_ENUM.containing_type = _BUNDLEDUPDATERESULT
_BUNDLEDUPDATESTATUS.fields_by_name['state'].enum_type = _BUNDLEDUPDATESTATE_ENUM
_BUNDLEDUPDATESTATUS.fields_by_name['result'].enum_type = _BUNDLEDUPDATERESULT_ENUM
_BUNDLEDUPDATESTATUS.fields_by_name['extended_status'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_BUNDLEDUPDATESTATUS.oneofs_by_name['_result'].fields.append(
  _BUNDLEDUPDATESTATUS.fields_by_name['result'])
_BUNDLEDUPDATESTATUS.fields_by_name['result'].containing_oneof = _BUNDLEDUPDATESTATUS.oneofs_by_name['_result']
_BUNDLEDUPDATESTATUS.oneofs_by_name['_current_state_progress_hundreth_percent'].fields.append(
  _BUNDLEDUPDATESTATUS.fields_by_name['current_state_progress_hundreth_percent'])
_BUNDLEDUPDATESTATUS.fields_by_name['current_state_progress_hundreth_percent'].containing_oneof = _BUNDLEDUPDATESTATUS.oneofs_by_name['_current_state_progress_hundreth_percent']
_BUNDLEDUPDATESTATUS.oneofs_by_name['_transfer_id'].fields.append(
  _BUNDLEDUPDATESTATUS.fields_by_name['transfer_id'])
_BUNDLEDUPDATESTATUS.fields_by_name['transfer_id'].containing_oneof = _BUNDLEDUPDATESTATUS.oneofs_by_name['_transfer_id']
_BUNDLEDUPDATESTATUS.oneofs_by_name['_bundle_filename'].fields.append(
  _BUNDLEDUPDATESTATUS.fields_by_name['bundle_filename'])
_BUNDLEDUPDATESTATUS.fields_by_name['bundle_filename'].containing_oneof = _BUNDLEDUPDATESTATUS.oneofs_by_name['_bundle_filename']
_BUNDLEDUPDATESTATUS.oneofs_by_name['_note'].fields.append(
  _BUNDLEDUPDATESTATUS.fields_by_name['note'])
_BUNDLEDUPDATESTATUS.fields_by_name['note'].containing_oneof = _BUNDLEDUPDATESTATUS.oneofs_by_name['_note']
_BUNDLEDUPDATESTATUS.oneofs_by_name['_extended_status'].fields.append(
  _BUNDLEDUPDATESTATUS.fields_by_name['extended_status'])
_BUNDLEDUPDATESTATUS.fields_by_name['extended_status'].containing_oneof = _BUNDLEDUPDATESTATUS.oneofs_by_name['_extended_status']
_STARTREQUEST.oneofs_by_name['_bundle_filename'].fields.append(
  _STARTREQUEST.fields_by_name['bundle_filename'])
_STARTREQUEST.fields_by_name['bundle_filename'].containing_oneof = _STARTREQUEST.oneofs_by_name['_bundle_filename']
DESCRIPTOR.message_types_by_name['BundledUpdateState'] = _BUNDLEDUPDATESTATE
DESCRIPTOR.message_types_by_name['BundledUpdateResult'] = _BUNDLEDUPDATERESULT
DESCRIPTOR.message_types_by_name['BundledUpdateStatus'] = _BUNDLEDUPDATESTATUS
DESCRIPTOR.message_types_by_name['StartRequest'] = _STARTREQUEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

BundledUpdateState = _reflection.GeneratedProtocolMessageType('BundledUpdateState', (_message.Message,), {
  'DESCRIPTOR' : _BUNDLEDUPDATESTATE,
  '__module__' : 'pw_software_update.bundled_update_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.BundledUpdateState)
  })
_sym_db.RegisterMessage(BundledUpdateState)

BundledUpdateResult = _reflection.GeneratedProtocolMessageType('BundledUpdateResult', (_message.Message,), {
  'DESCRIPTOR' : _BUNDLEDUPDATERESULT,
  '__module__' : 'pw_software_update.bundled_update_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.BundledUpdateResult)
  })
_sym_db.RegisterMessage(BundledUpdateResult)

BundledUpdateStatus = _reflection.GeneratedProtocolMessageType('BundledUpdateStatus', (_message.Message,), {
  'DESCRIPTOR' : _BUNDLEDUPDATESTATUS,
  '__module__' : 'pw_software_update.bundled_update_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.BundledUpdateStatus)
  })
_sym_db.RegisterMessage(BundledUpdateStatus)

StartRequest = _reflection.GeneratedProtocolMessageType('StartRequest', (_message.Message,), {
  'DESCRIPTOR' : _STARTREQUEST,
  '__module__' : 'pw_software_update.bundled_update_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.StartRequest)
  })
_sym_db.RegisterMessage(StartRequest)


_BUNDLEDUPDATESTATUS.fields_by_name['note']._options = None

_BUNDLEDUPDATE = _descriptor.ServiceDescriptor(
  name='BundledUpdate',
  full_name='pw.software_update.BundledUpdate',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=995,
  serialized_end=1528,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetStatus',
    full_name='pw.software_update.BundledUpdate.GetStatus',
    index=0,
    containing_service=None,
    input_type=pw__protobuf__protos_dot_common__pb2._EMPTY,
    output_type=_BUNDLEDUPDATESTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Start',
    full_name='pw.software_update.BundledUpdate.Start',
    index=1,
    containing_service=None,
    input_type=_STARTREQUEST,
    output_type=_BUNDLEDUPDATESTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SetTransferred',
    full_name='pw.software_update.BundledUpdate.SetTransferred',
    index=2,
    containing_service=None,
    input_type=pw__protobuf__protos_dot_common__pb2._EMPTY,
    output_type=_BUNDLEDUPDATESTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Verify',
    full_name='pw.software_update.BundledUpdate.Verify',
    index=3,
    containing_service=None,
    input_type=pw__protobuf__protos_dot_common__pb2._EMPTY,
    output_type=_BUNDLEDUPDATESTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Apply',
    full_name='pw.software_update.BundledUpdate.Apply',
    index=4,
    containing_service=None,
    input_type=pw__protobuf__protos_dot_common__pb2._EMPTY,
    output_type=_BUNDLEDUPDATESTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Abort',
    full_name='pw.software_update.BundledUpdate.Abort',
    index=5,
    containing_service=None,
    input_type=pw__protobuf__protos_dot_common__pb2._EMPTY,
    output_type=_BUNDLEDUPDATESTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Reset',
    full_name='pw.software_update.BundledUpdate.Reset',
    index=6,
    containing_service=None,
    input_type=pw__protobuf__protos_dot_common__pb2._EMPTY,
    output_type=_BUNDLEDUPDATESTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_BUNDLEDUPDATE)

DESCRIPTOR.services_by_name['BundledUpdate'] = _BUNDLEDUPDATE

# @@protoc_insertion_point(module_scope)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""UI Color Styles for ConsoleApp."""

import logging
from dataclasses import dataclass

from prompt_toolkit.formatted_text import StyleAndTextTuples
from prompt_toolkit.formatted_text.base import OneStyleAndTextTuple
from prompt_toolkit.styles import Style
from prompt_toolkit.filters import has_focus

_LOG = logging.getLogger(__package__)


@dataclass
class HighContrastDarkColors:
    # pylint: disable=too-many-instance-attributes
    default_bg = '#100f10'
    default_fg = '#ffffff'

    dim_bg = '#000000'
    dim_fg = '#e0e6f0'

    button_active_bg = '#4e4e4e'
    button_inactive_bg = '#323232'

    active_bg = '#323232'
    active_fg = '#f4f4f4'

    inactive_bg = '#1e1e1e'
    inactive_fg = '#bfc0c4'

    line_highlight_bg = '#2f2f2f'
    selected_line_bg = '#4e4e4e'
    dialog_bg = '#3c3c3c'

    red_accent = '#ffc0bf'
    orange_accent = '#f5ca80'
    yellow_accent = '#eedc82'
    green_accent = '#88ef88'
    cyan_accent = '#60e7e0'
    blue_accent = '#92d9ff'
    purple_accent = '#cfcaff'
    magenta_accent = '#ffb8ff'


@dataclass
class DarkColors:
    # pylint: disable=too-many-instance-attributes
    default_bg = '#2e2e2e'
    default_fg = '#eeeeee'

    dim_bg = '#262626'
    dim_fg = '#dfdfdf'

    button_active_bg = '#626262'
    button_inactive_bg = '#525252'

    active_bg = '#525252'
    active_fg = '#dfdfdf'

    inactive_bg = '#3f3f3f'
    inactive_fg = '#bfbfbf'

    line_highlight_bg = '#525252'
    selected_line_bg = '#626262'
    dialog_bg = '#3c3c3c'

    red_accent = '#ff6c6b'
    orange_accent = '#da8548'
    yellow_accent = '#ffcc66'
    green_accent = '#98be65'
    cyan_accent = '#66cccc'
    blue_accent = '#6699cc'
    purple_accent = '#a9a1e1'
    magenta_accent = '#c678dd'


@dataclass
class NordColors:
    # pylint: disable=too-many-instance-attributes
    default_bg = '#2e3440'
    default_fg = '#eceff4'

    dim_bg = '#272c36'
    dim_fg = '#e5e9f0'

    button_active_bg = '#4c566a'
    button_inactive_bg = '#434c5e'

    active_bg = '#434c5e'
    active_fg = '#eceff4'

    inactive_bg = '#373e4c'
    inactive_fg = '#d8dee9'

    line_highlight_bg = '#191c25'
    selected_line_bg = '#4c566a'
    dialog_bg = '#2c333f'

    red_accent = '#bf616a'
    orange_accent = '#d08770'
    yellow_accent = '#ebcb8b'
    green_accent = '#a3be8c'
    cyan_accent = '#88c0d0'
    blue_accent = '#81a1c1'
    purple_accent = '#a9a1e1'
    magenta_accent = '#b48ead'


@dataclass
class NordLightColors:
    # pylint: disable=too-many-instance-attributes
    default_bg = '#e5e9f0'
    default_fg = '#3b4252'
    dim_bg = '#d8dee9'
    dim_fg = '#2e3440'
    button_active_bg = '#aebacf'
    button_inactive_bg = '#b8c5db'
    active_bg = '#b8c5db'
    active_fg = '#3b4252'
    inactive_bg = '#c2d0e7'
    inactive_fg = '#60728c'
    line_highlight_bg = '#f0f4fc'
    selected_line_bg = '#f0f4fc'
    dialog_bg = '#d8dee9'

    red_accent = '#99324b'
    orange_accent = '#ac4426'
    yellow_accent = '#9a7500'
    green_accent = '#4f894c'
    cyan_accent = '#398eac'
    blue_accent = '#3b6ea8'
    purple_accent = '#842879'
    magenta_accent = '#97365b'


@dataclass
class MoonlightColors:
    # pylint: disable=too-many-instance-attributes
    default_bg = '#212337'
    default_fg = '#c8d3f5'
    dim_bg = '#191a2a'
    dim_fg = '#b4c2f0'
    button_active_bg = '#444a73'
    button_inactive_bg = '#2f334d'
    active_bg = '#2f334d'
    active_fg = '#c8d3f5'
    inactive_bg = '#222436'
    inactive_fg = '#a9b8e8'
    line_highlight_bg = '#383e5c'
    selected_line_bg = '#444a73'
    dialog_bg = '#1e2030'

    red_accent = '#d95468'
    orange_accent = '#d98e48'
    yellow_accent = '#ebbf83'
    green_accent = '#8bd49c'
    cyan_accent = '#70e1e8'
    blue_accent = '#5ec4ff'
    purple_accent = '#b62d65'
    magenta_accent = '#e27e8d'


@dataclass
class AnsiTerm:
    # pylint: disable=too-many-instance-attributes
    default_bg = 'default'
    default_fg = 'default'

    dim_bg = 'default'
    dim_fg = 'default'

    button_active_bg = 'default underline'
    button_inactive_bg = 'default'

    active_bg = 'default'
    active_fg = 'default'

    inactive_bg = 'default'
    inactive_fg = 'default'

    line_highlight_bg = 'ansidarkgray white'
    selected_line_bg = 'default reverse'
    dialog_bg = 'default'

    red_accent = 'ansired'
    orange_accent = 'orange'
    yellow_accent = 'ansiyellow'
    green_accent = 'ansigreen'
    cyan_accent = 'ansicyan'
    blue_accent = 'ansiblue'
    purple_accent = 'ansipurple'
    magenta_accent = 'ansimagenta'


_THEME_NAME_MAPPING = {
    'moonlight': MoonlightColors(),
    'nord': NordColors(),
    'nord-light': NordLightColors(),
    'dark': DarkColors(),
    'high-contrast-dark': HighContrastDarkColors(),
    'ansi': AnsiTerm(),
}


def get_theme_colors(theme_name=''):
    theme = _THEME_NAME_MAPPING.get(theme_name, DarkColors())
    return theme


def generate_styles(theme_name='dark'):
    """Return prompt_toolkit styles for the given theme name."""
    # Use DarkColors() if name not found.
    theme = _THEME_NAME_MAPPING.get(theme_name, DarkColors())

    pw_console_styles = {
        # Default text and background.
        'default': 'bg:{} {}'.format(theme.default_bg, theme.default_fg),
        # Dim inactive panes.
        'pane_inactive': 'bg:{} {}'.format(theme.dim_bg, theme.dim_fg),
        # Use default for active panes.
        'pane_active': 'bg:{} {}'.format(theme.default_bg, theme.default_fg),
        # Brighten active pane toolbars.
        'toolbar_active': 'bg:{} {}'.format(theme.active_bg, theme.active_fg),
        'toolbar_inactive': 'bg:{} {}'.format(
            theme.inactive_bg, theme.inactive_fg
        ),
        # Dimmer toolbar.
        'toolbar_dim_active': 'bg:{} {}'.format(
            theme.active_bg, theme.active_fg
        ),
        'toolbar_dim_inactive': 'bg:{} {}'.format(
            theme.default_bg, theme.inactive_fg
        ),
        # Used for pane titles
        'toolbar_accent': theme.cyan_accent,
        'toolbar-button-decoration': '{}'.format(theme.cyan_accent),
        'toolbar-setting-active': 'bg:{} {}'.format(
            theme.green_accent,
            theme.active_bg,
        ),
        'toolbar-button-active': 'bg:{}'.format(theme.button_active_bg),
        'toolbar-button-inactive': 'bg:{}'.format(theme.button_inactive_bg),
        # prompt_toolkit scrollbar styles:
        'scrollbar.background': 'bg:{} {}'.format(
            theme.default_bg, theme.default_fg
        ),
        # Scrollbar handle, bg is the bar color.
        'scrollbar.button': 'bg:{} {}'.format(
            theme.purple_accent, theme.default_bg
        ),
        'scrollbar.arrow': 'bg:{} {}'.format(
            theme.default_bg, theme.blue_accent
        ),
        # Unstyled scrollbar classes:
        # 'scrollbar.start'
        # 'scrollbar.end'
        # Top menu bar styles
        'menu-bar': 'bg:{} {}'.format(theme.inactive_bg, theme.inactive_fg),
        'menu-bar.selected-item': 'bg:{} {}'.format(
            theme.blue_accent, theme.inactive_bg
        ),
        # Menu background
        'menu': 'bg:{} {}'.format(theme.dialog_bg, theme.dim_fg),
        # Menu item separator
        'menu-border': theme.magenta_accent,
        # Top bar logo + keyboard shortcuts
        'logo': '{} bold'.format(theme.magenta_accent),
        'keybind': '{} bold'.format(theme.purple_accent),
        'keyhelp': theme.dim_fg,
        # Help window styles
        'help_window_content': 'bg:{} {}'.format(theme.dialog_bg, theme.dim_fg),
        'frame.border': 'bg:{} {}'.format(theme.dialog_bg, theme.purple_accent),
        'pane_indicator_active': 'bg:{}'.format(theme.magenta_accent),
        'pane_indicator_inactive': 'bg:{}'.format(theme.inactive_bg),
        'pane_title_active': '{} bold'.format(theme.magenta_accent),
        'pane_title_inactive': '{}'.format(theme.purple_accent),
        'window-tab-active': 'bg:{} {}'.format(
            theme.active_bg, theme.cyan_accent
        ),
        'window-tab-inactive': 'bg:{} {}'.format(
            theme.inactive_bg, theme.inactive_fg
        ),
        'pane_separator': 'bg:{} {}'.format(
            theme.default_bg, theme.purple_accent
        ),
        # Search matches
        'search': 'bg:{} {}'.format(theme.cyan_accent, theme.default_bg),
        'search.current': 'bg:{} {}'.format(
            theme.cyan_accent, theme.default_bg
        ),
        # Highlighted line styles
        'selected-log-line': 'bg:{}'.format(theme.line_highlight_bg),
        'marked-log-line': 'bg:{}'.format(theme.selected_line_bg),
        'cursor-line': 'bg:{} nounderline'.format(theme.line_highlight_bg),
        # Messages like 'Window too small'
        'warning-text': 'bg:{} {}'.format(
            theme.default_bg, theme.yellow_accent
        ),
        'log-time': 'bg:{} {}'.format(theme.default_fg, theme.default_bg),
        # Apply foreground only for level and column values. This way the text
        # can inherit the background color of the parent window pane or line
        # selection.
        'log-level-{}'.format(logging.CRITICAL): '{} bold'.format(
            theme.red_accent
        ),
        'log-level-{}'.format(logging.ERROR): '{}'.format(theme.red_accent),
        'log-level-{}'.format(logging.WARNING): '{}'.format(
            theme.yellow_accent
        ),
        'log-level-{}'.format(logging.INFO): '{}'.format(theme.purple_accent),
        'log-level-{}'.format(logging.DEBUG): '{}'.format(theme.blue_accent),
        'log-table-column-0': '{}'.format(theme.cyan_accent),
        'log-table-column-1': '{}'.format(theme.green_accent),
        'log-table-column-2': '{}'.format(theme.yellow_accent),
        'log-table-column-3': '{}'.format(theme.magenta_accent),
        'log-table-column-4': '{}'.format(theme.purple_accent),
        'log-table-column-5': '{}'.format(theme.blue_accent),
        'log-table-column-6': '{}'.format(theme.orange_accent),
        'log-table-column-7': '{}'.format(theme.red_accent),
        'search-bar': 'bg:{}'.format(theme.inactive_bg),
        'search-bar-title': 'bg:{} {}'.format(
            theme.cyan_accent, theme.default_bg
        ),
        'search-bar-setting': '{}'.format(theme.cyan_accent),
        'search-bar-border': 'bg:{} {}'.format(
            theme.inactive_bg, theme.cyan_accent
        ),
        'search-match-count-dialog': 'bg:{}'.format(theme.inactive_bg),
        'search-match-count-dialog-title': '{}'.format(theme.cyan_accent),
        'search-match-count-dialog-default-fg': '{}'.format(theme.default_fg),
        'search-match-count-dialog-border': 'bg:{} {}'.format(
            theme.inactive_bg, theme.cyan_accent
        ),
        'filter-bar': 'bg:{}'.format(theme.inactive_bg),
        'filter-bar-title': 'bg:{} {}'.format(
            theme.red_accent, theme.default_bg
        ),
        'filter-bar-setting': '{}'.format(theme.cyan_accent),
        'filter-bar-delete': '{}'.format(theme.red_accent),
        'filter-bar-delimiter': '{}'.format(theme.purple_accent),
        'saveas-dialog': 'bg:{}'.format(theme.inactive_bg),
        'saveas-dialog-title': 'bg:{} {}'.format(
            theme.inactive_bg, theme.default_fg
        ),
        'saveas-dialog-setting': '{}'.format(theme.cyan_accent),
        'saveas-dialog-border': 'bg:{} {}'.format(
            theme.inactive_bg, theme.cyan_accent
        ),
        'selection-dialog': 'bg:{}'.format(theme.inactive_bg),
        'selection-dialog-title': '{}'.format(theme.yellow_accent),
        'selection-dialog-default-fg': '{}'.format(theme.default_fg),
        'selection-dialog-action-bg': 'bg:{}'.format(theme.yellow_accent),
        'selection-dialog-action-fg': '{}'.format(theme.button_inactive_bg),
        'selection-dialog-border': 'bg:{} {}'.format(
            theme.inactive_bg, theme.yellow_accent
        ),
        'quit-dialog': 'bg:{}'.format(theme.inactive_bg),
        'quit-dialog-border': 'bg:{} {}'.format(
            theme.inactive_bg, theme.red_accent
        ),
        'command-runner': 'bg:{}'.format(theme.inactive_bg),
        'command-runner-title': 'bg:{} {}'.format(
            theme.inactive_bg, theme.default_fg
        ),
        'command-runner-setting': '{}'.format(theme.purple_accent),
        'command-runner-border': 'bg:{} {}'.format(
            theme.inactive_bg, theme.purple_accent
        ),
        'command-runner-selected-item': 'bg:{}'.format(theme.selected_line_bg),
        'command-runner-fuzzy-highlight-0': '{}'.format(theme.blue_accent),
        'command-runner-fuzzy-highlight-1': '{}'.format(theme.cyan_accent),
        'command-runner-fuzzy-highlight-2': '{}'.format(theme.green_accent),
        'command-runner-fuzzy-highlight-3': '{}'.format(theme.yellow_accent),
        'command-runner-fuzzy-highlight-4': '{}'.format(theme.orange_accent),
        'command-runner-fuzzy-highlight-5': '{}'.format(theme.red_accent),
        # Progress Bar Styles
        # Entire set of ProgressBars - no title is used in pw_console
        'title': '',
        # Actual bar title
        'label': 'bold',
        'percentage': '{}'.format(theme.green_accent),
        'bar': '{}'.format(theme.magenta_accent),
        # Filled part of the bar
        'bar-a': '{} bold'.format(theme.cyan_accent),
        # End of current progress
        'bar-b': '{} bold'.format(theme.purple_accent),
        # Empty part of the bar
        'bar-c': '',
        # current/total counts
        'current': '{}'.format(theme.cyan_accent),
        'total': '{}'.format(theme.cyan_accent),
        'time-elapsed': '{}'.format(theme.purple_accent),
        'time-left': '{}'.format(theme.magenta_accent),
        # Named theme color classes for use in user plugins.
        'theme-fg-red': '{}'.format(theme.red_accent),
        'theme-fg-orange': '{}'.format(theme.orange_accent),
        'theme-fg-yellow': '{}'.format(theme.yellow_accent),
        'theme-fg-green': '{}'.format(theme.green_accent),
        'theme-fg-cyan': '{}'.format(theme.cyan_accent),
        'theme-fg-blue': '{}'.format(theme.blue_accent),
        'theme-fg-purple': '{}'.format(theme.purple_accent),
        'theme-fg-magenta': '{}'.format(theme.magenta_accent),
        'theme-bg-red': 'bg:{}'.format(theme.red_accent),
        'theme-bg-orange': 'bg:{}'.format(theme.orange_accent),
        'theme-bg-yellow': 'bg:{}'.format(theme.yellow_accent),
        'theme-bg-green': 'bg:{}'.format(theme.green_accent),
        'theme-bg-cyan': 'bg:{}'.format(theme.cyan_accent),
        'theme-bg-blue': 'bg:{}'.format(theme.blue_accent),
        'theme-bg-purple': 'bg:{}'.format(theme.purple_accent),
        'theme-bg-magenta': 'bg:{}'.format(theme.magenta_accent),
        'theme-bg-active': 'bg:{}'.format(theme.active_bg),
        'theme-fg-active': '{}'.format(theme.active_fg),
        'theme-bg-inactive': 'bg:{}'.format(theme.inactive_bg),
        'theme-fg-inactive': '{}'.format(theme.inactive_fg),
        'theme-fg-default': '{}'.format(theme.default_fg),
        'theme-bg-default': 'bg:{}'.format(theme.default_bg),
        'theme-fg-dim': '{}'.format(theme.dim_fg),
        'theme-bg-dim': 'bg:{}'.format(theme.dim_bg),
        'theme-bg-dialog': 'bg:{}'.format(theme.dialog_bg),
        'theme-bg-line-highlight': 'bg:{}'.format(theme.line_highlight_bg),
        'theme-bg-button-active': 'bg:{}'.format(theme.button_active_bg),
        'theme-bg-button-inactive': 'bg:{}'.format(theme.button_inactive_bg),
    }

    return Style.from_dict(pw_console_styles)


def get_toolbar_style(pt_container, dim=False) -> str:
    """Return the style class for a toolbar if pt_container is in focus."""
    if has_focus(pt_container.__pt_container__())():
        return 'class:toolbar_dim_active' if dim else 'class:toolbar_active'
    return 'class:toolbar_dim_inactive' if dim else 'class:toolbar_inactive'


def get_button_style(pt_container) -> str:
    """Return the style class for a toolbar if pt_container is in focus."""
    if has_focus(pt_container.__pt_container__())():
        return 'class:toolbar-button-active'
    return 'class:toolbar-button-inactive'


def get_pane_style(pt_container) -> str:
    """Return the style class for a pane title if pt_container is in focus."""
    if has_focus(pt_container.__pt_container__())():
        return 'class:pane_active'
    return 'class:pane_inactive'


def get_pane_indicator(
    pt_container, title, mouse_handler=None, hide_indicator=False
) -> StyleAndTextTuples:
    """Return formatted text for a pane indicator and title."""

    inactive_indicator: OneStyleAndTextTuple
    active_indicator: OneStyleAndTextTuple
    inactive_title: OneStyleAndTextTuple
    active_title: OneStyleAndTextTuple

    if mouse_handler:
        inactive_indicator = (
            'class:pane_indicator_inactive',
            ' ',
            mouse_handler,
        )
        active_indicator = ('class:pane_indicator_active', ' ', mouse_handler)
        inactive_title = ('class:pane_title_inactive', title, mouse_handler)
        active_title = ('class:pane_title_active', title, mouse_handler)
    else:
        inactive_indicator = ('class:pane_indicator_inactive', ' ')
        active_indicator = ('class:pane_indicator_active', ' ')
        inactive_title = ('class:pane_title_inactive', title)
        active_title = ('class:pane_title_active', title)

    fragments: StyleAndTextTuples = []
    if has_focus(pt_container.__pt_container__())():
        if not hide_indicator:
            fragments.append(active_indicator)
        fragments.append(active_title)
    else:
        if not hide_indicator:
            fragments.append(inactive_indicator)
        fragments.append(inactive_title)
    return fragments



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Tools for working with tokenized logs."""

from dataclasses import dataclass, asdict
import re
from typing import Dict, Mapping, Iterator


def _mask(value: int, start: int, count: int) -> int:
    mask = (1 << count) - 1
    return (value & (mask << start)) >> start


@dataclass
class Metadata:
    """Parses the metadata payload used by pw_log_tokenized."""

    value: int
    log_bits: int = 3
    line_bits: int = 11
    flag_bits: int = 2
    module_bits: int = 16

    def __post_init__(self):
        self.log_level = _mask(self.value, 0, self.log_bits)
        self.line = _mask(self.value, self.log_bits, self.line_bits)
        self.flags = _mask(
            self.value, self.log_bits + self.line_bits, self.flag_bits
        )
        self.module_token = _mask(
            self.value,
            self.log_bits + self.line_bits + self.flag_bits,
            self.module_bits,
        )

    def __iter__(self):
        return iter(asdict(self).items())

    def __dict__(self):
        return asdict(self)


class FormatStringWithMetadata:
    """Parses metadata from a log format string with metadata fields."""

    _FIELD_KEY = re.compile(r'■([a-zA-Z]\w*)♦', flags=re.ASCII)

    def __init__(self, string: str) -> None:
        self.raw_string = string
        self.fields: Dict[str, str] = {}

        # Only look for fields if the raw string starts with one.
        if self._FIELD_KEY.match(self.raw_string):
            fields = self._FIELD_KEY.split(self.raw_string)[1:]
            for name, value in zip(fields[::2], fields[1::2]):
                self.fields[name] = value

    @property
    def message(self) -> str:
        """Displays the msg field or the whole string if it is not present."""
        return self.fields.get('msg', self.raw_string)

    @property
    def module(self) -> str:
        return self.fields.get('module', '')

    @property
    def file(self) -> str:
        return self.fields.get('file', '')

    def __repr__(self) -> str:
        return self.message



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2020 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _TestCaseResult:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _TestCaseResultEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_TestCaseResult.ValueType], builtins.type):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    SUCCESS: _TestCaseResult.ValueType  # 0
    FAILURE: _TestCaseResult.ValueType  # 1
    SKIPPED: _TestCaseResult.ValueType  # 2

class TestCaseResult(_TestCaseResult, metaclass=_TestCaseResultEnumTypeWrapper): ...

SUCCESS: TestCaseResult.ValueType  # 0
FAILURE: TestCaseResult.ValueType  # 1
SKIPPED: TestCaseResult.ValueType  # 2
global___TestCaseResult = TestCaseResult

class TestCaseDescriptor(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SUITE_NAME_FIELD_NUMBER: builtins.int
    TEST_NAME_FIELD_NUMBER: builtins.int
    FILE_NAME_FIELD_NUMBER: builtins.int
    suite_name: builtins.str
    """Name of the test suite to which this test case belongs."""
    test_name: builtins.str
    """Name of the test case."""
    file_name: builtins.str
    """Path to the file in which the test case is defined."""
    def __init__(
        self,
        *,
        suite_name: builtins.str = ...,
        test_name: builtins.str = ...,
        file_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["file_name", b"file_name", "suite_name", b"suite_name", "test_name", b"test_name"]) -> None: ...

global___TestCaseDescriptor = TestCaseDescriptor

class TestCaseExpectation(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EXPRESSION_FIELD_NUMBER: builtins.int
    EVALUATED_EXPRESSION_FIELD_NUMBER: builtins.int
    LINE_NUMBER_FIELD_NUMBER: builtins.int
    SUCCESS_FIELD_NUMBER: builtins.int
    expression: builtins.str
    """The source code for the expression which was run."""
    evaluated_expression: builtins.str
    """The expression with arguments evaluated."""
    line_number: builtins.int
    """Line number at which the expectation is located."""
    success: builtins.bool
    """Whether the expectation succeeded."""
    def __init__(
        self,
        *,
        expression: builtins.str = ...,
        evaluated_expression: builtins.str = ...,
        line_number: builtins.int = ...,
        success: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["evaluated_expression", b"evaluated_expression", "expression", b"expression", "line_number", b"line_number", "success", b"success"]) -> None: ...

global___TestCaseExpectation = TestCaseExpectation

class TestRunStart(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___TestRunStart = TestRunStart

class TestRunEnd(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PASSED_FIELD_NUMBER: builtins.int
    FAILED_FIELD_NUMBER: builtins.int
    SKIPPED_FIELD_NUMBER: builtins.int
    DISABLED_FIELD_NUMBER: builtins.int
    passed: builtins.int
    failed: builtins.int
    skipped: builtins.int
    disabled: builtins.int
    def __init__(
        self,
        *,
        passed: builtins.int = ...,
        failed: builtins.int = ...,
        skipped: builtins.int = ...,
        disabled: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["disabled", b"disabled", "failed", b"failed", "passed", b"passed", "skipped", b"skipped"]) -> None: ...

global___TestRunEnd = TestRunEnd

class Event(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TEST_RUN_START_FIELD_NUMBER: builtins.int
    TEST_RUN_END_FIELD_NUMBER: builtins.int
    TEST_CASE_START_FIELD_NUMBER: builtins.int
    TEST_CASE_END_FIELD_NUMBER: builtins.int
    TEST_CASE_DISABLED_FIELD_NUMBER: builtins.int
    TEST_CASE_EXPECTATION_FIELD_NUMBER: builtins.int
    @property
    def test_run_start(self) -> global___TestRunStart:
        """Unit test run has started."""
    @property
    def test_run_end(self) -> global___TestRunEnd:
        """Unit test run has ended."""
    @property
    def test_case_start(self) -> global___TestCaseDescriptor:
        """Start of a test case."""
    test_case_end: global___TestCaseResult.ValueType
    """End of a test case."""
    @property
    def test_case_disabled(self) -> global___TestCaseDescriptor:
        """Encountered a disabled test case."""
    @property
    def test_case_expectation(self) -> global___TestCaseExpectation:
        """Expectation statement within a test case."""
    def __init__(
        self,
        *,
        test_run_start: global___TestRunStart | None = ...,
        test_run_end: global___TestRunEnd | None = ...,
        test_case_start: global___TestCaseDescriptor | None = ...,
        test_case_end: global___TestCaseResult.ValueType = ...,
        test_case_disabled: global___TestCaseDescriptor | None = ...,
        test_case_expectation: global___TestCaseExpectation | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["test_case_disabled", b"test_case_disabled", "test_case_end", b"test_case_end", "test_case_expectation", b"test_case_expectation", "test_case_start", b"test_case_start", "test_run_end", b"test_run_end", "test_run_start", b"test_run_start", "type", b"type"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["test_case_disabled", b"test_case_disabled", "test_case_end", b"test_case_end", "test_case_expectation", b"test_case_expectation", "test_case_start", b"test_case_start", "test_run_end", b"test_run_end", "test_run_start", b"test_run_start", "type", b"type"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["type", b"type"]) -> typing_extensions.Literal["test_run_start", "test_run_end", "test_case_start", "test_case_end", "test_case_disabled", "test_case_expectation"] | None: ...

global___Event = Event

class TestRunRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REPORT_PASSED_EXPECTATIONS_FIELD_NUMBER: builtins.int
    TEST_SUITE_FIELD_NUMBER: builtins.int
    report_passed_expectations: builtins.bool
    """Whether to send expectation events for successful checks."""
    @property
    def test_suite(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Optional list of test suites to run."""
    def __init__(
        self,
        *,
        report_passed_expectations: builtins.bool = ...,
        test_suite: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["report_passed_expectations", b"report_passed_expectations", "test_suite", b"test_suite"]) -> None: ...

global___TestRunRequest = TestRunRequest



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Install and check status of Emboss."""

import pathlib
from typing import Sequence

import pw_package.git_repo
import pw_package.package_manager


class Emboss(pw_package.git_repo.GitRepo):
    """Install and check status of Emboss."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            name="emboss",
            url="".join(
                [
                    "https://fuchsia.googlesource.com",
                    "/third_party/github.com/google/emboss",
                ]
            ),
            commit="b02439aeecb9cc8621c7aea9fc206f65da45c9be",
            **kwargs,
        )

    def info(self, path: pathlib.Path) -> Sequence[str]:
        return (
            f"{self.name} installed in: {path}",
            "Enable by running 'gn args out' and adding this line:",
            f'  dir_pw_third_party_emboss = "{path}"',
        )


pw_package.package_manager.register(Emboss)



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Cortex-M architecture related constants."""

import collections

# Cortex-M (ARMv7-M + ARMv8-M) related constants.
# These values are from the ARMv7-M Architecture Reference Manual DDI 0403E.b
# and ARMv8-M Architecture Reference Manual DDI 0553A.e.
# https =//static.docs.arm.com/ddi0403/e/DDI0403E_B_armv7m_arm.pdf
# https =//static.docs.arm.com/ddi0553/a/DDI0553A_e_armv8m_arm.pdf

# Exception ISR number. (ARMv7-M Section B1.5.2)
# When the ISR number (accessible from ICSR and PSR) is zero, it indicates the
# core is in thread mode.
PW_CORTEX_M_THREAD_MODE_ISR_NUM = 0x0
PW_CORTEX_M_NMI_ISR_NUM = 0x2
PW_CORTEX_M_HARD_FAULT_ISR_NUM = 0x3
PW_CORTEX_M_MEM_FAULT_ISR_NUM = 0x4
PW_CORTEX_M_BUS_FAULT_ISR_NUM = 0x5
PW_CORTEX_M_USAGE_FAULT_ISR_NUM = 0x6

# Masks for Interrupt Control and State Register ICSR (ARMv7-M Section B3.2.4)
PW_CORTEX_M_ICSR_VECTACTIVE_MASK = (1 << 9) - 1

# Masks for individual bits of HFSR. (ARMv7-M Section B3.2.16)
PW_CORTEX_M_HFSR_FORCED_MASK = 0x1 << 30

# Masks for different sections of CFSR. (ARMv7-M Section B3.2.15)
PW_CORTEX_M_CFSR_MEM_FAULT_MASK = 0x000000FF
PW_CORTEX_M_CFSR_BUS_FAULT_MASK = 0x0000FF00
PW_CORTEX_M_CFSR_USAGE_FAULT_MASK = 0xFFFF0000

# Masks for individual bits of CFSR. (ARMv7-M Section B3.2.15)
# Memory faults (MemManage Status Register) =
PW_CORTEX_M_CFSR_MEM_FAULT_START = 0x1
PW_CORTEX_M_CFSR_IACCVIOL_MASK = PW_CORTEX_M_CFSR_MEM_FAULT_START << 0
PW_CORTEX_M_CFSR_DACCVIOL_MASK = PW_CORTEX_M_CFSR_MEM_FAULT_START << 1
PW_CORTEX_M_CFSR_MUNSTKERR_MASK = PW_CORTEX_M_CFSR_MEM_FAULT_START << 3
PW_CORTEX_M_CFSR_MSTKERR_MASK = PW_CORTEX_M_CFSR_MEM_FAULT_START << 4
PW_CORTEX_M_CFSR_MLSPERR_MASK = PW_CORTEX_M_CFSR_MEM_FAULT_START << 5
PW_CORTEX_M_CFSR_MMARVALID_MASK = PW_CORTEX_M_CFSR_MEM_FAULT_START << 7
# Bus faults (BusFault Status Register) =
PW_CORTEX_M_CFSR_BUS_FAULT_START = 0x1 << 8
PW_CORTEX_M_CFSR_IBUSERR_MASK = PW_CORTEX_M_CFSR_BUS_FAULT_START << 0
PW_CORTEX_M_CFSR_PRECISERR_MASK = PW_CORTEX_M_CFSR_BUS_FAULT_START << 1
PW_CORTEX_M_CFSR_IMPRECISERR_MASK = PW_CORTEX_M_CFSR_BUS_FAULT_START << 2
PW_CORTEX_M_CFSR_UNSTKERR_MASK = PW_CORTEX_M_CFSR_BUS_FAULT_START << 3
PW_CORTEX_M_CFSR_STKERR_MASK = PW_CORTEX_M_CFSR_BUS_FAULT_START << 4
PW_CORTEX_M_CFSR_LSPERR_MASK = PW_CORTEX_M_CFSR_BUS_FAULT_START << 5
PW_CORTEX_M_CFSR_BFARVALID_MASK = PW_CORTEX_M_CFSR_BUS_FAULT_START << 7
# Usage faults (UsageFault Status Register) =
PW_CORTEX_M_CFSR_USAGE_FAULT_START = 0x1 << 16
PW_CORTEX_M_CFSR_UNDEFINSTR_MASK = PW_CORTEX_M_CFSR_USAGE_FAULT_START << 0
PW_CORTEX_M_CFSR_INVSTATE_MASK = PW_CORTEX_M_CFSR_USAGE_FAULT_START << 1
PW_CORTEX_M_CFSR_INVPC_MASK = PW_CORTEX_M_CFSR_USAGE_FAULT_START << 2
PW_CORTEX_M_CFSR_NOCP_MASK = PW_CORTEX_M_CFSR_USAGE_FAULT_START << 3
PW_CORTEX_M_CFSR_STKOF_MASK = PW_CORTEX_M_CFSR_USAGE_FAULT_START << 4
PW_CORTEX_M_CFSR_UNALIGNED_MASK = PW_CORTEX_M_CFSR_USAGE_FAULT_START << 8
PW_CORTEX_M_CFSR_DIVBYZERO_MASK = PW_CORTEX_M_CFSR_USAGE_FAULT_START << 9

# TODO(amontanez): We could probably make a whole module on bit field handling
# in python.
BitField = collections.namedtuple(
    'BitField', ['name', 'bit_mask', 'description', 'long_description']
)

# Information about faults from:
# * ARM Cortex-M4 Devices Generic User Guide 4.3.10
# * ARM Cortex-M33 Devices Generic User Guide 4.2.11

PW_CORTEX_M_CFSR_BIT_FIELDS = [
    BitField(
        'IACCVIOL',
        PW_CORTEX_M_CFSR_IACCVIOL_MASK,
        'Instruction access violation fault.',
        (
            'The processor attempted an instruction fetch from a location',
            'that does not permit execution. The PC value stacked for the',
            'exception return points to the faulting instruction.',
        ),
    ),
    BitField(
        'DACCVIOL',
        PW_CORTEX_M_CFSR_DACCVIOL_MASK,
        'Data access violation fault.',
        (
            'The processor attempted a load or store at a location that',
            'does not permit the operation. The PC value stacked for the',
            'exception return points to the faulting instruction. The',
            'processor has loaded the MMFAR with the address of the',
            'attempted access.',
        ),
    ),
    BitField(
        'MUNSTKERR',
        PW_CORTEX_M_CFSR_MUNSTKERR_MASK,
        'MemManage fault on unstacking for a return from exception.',
        (
            'Unstack for an exception return has caused one or more access',
            'violations. This fault is chained to the handler. This means',
            'that when this bit is 1, the original return stack is still',
            'present. The processor has not adjusted the SP from the',
            'failing return, and has not performed a new save. The',
            'processor has not written a fault address to the MMAR.',
        ),
    ),
    BitField(
        'MSTKERR',
        PW_CORTEX_M_CFSR_MSTKERR_MASK,
        'MemManage fault on stacking for exception entry.',
        (
            'When this bit is 1, the SP is still adjusted but the values',
            'in the context area on the stack might be incorrect. The',
            'processor has not written a fault address to the MMAR.',
        ),
    ),
    BitField(
        'MLSPERR',
        PW_CORTEX_M_CFSR_MLSPERR_MASK,
        'MemManage Fault during FPU lazy state preservation.',
        '',
    ),
    BitField(
        'MMARVALID',
        PW_CORTEX_M_CFSR_MMARVALID_MASK,
        'MMFAR register is valid.',
        '',
    ),
    BitField(
        'IBUSERR',
        PW_CORTEX_M_CFSR_IBUSERR_MASK,
        'Instruction bus error.',
        (
            'The processor attempted to issue an invalid instruction. It',
            'detects the instruction bus error on prefetching, but this',
            'flag is only set to 1 if it attempts to issue the faulting',
            'instruction. When this bit is set, the processor has not',
            'written a fault address to the BFAR.',
        ),
    ),
    BitField(
        'PRECISERR',
        PW_CORTEX_M_CFSR_PRECISERR_MASK,
        'Precise data bus error.',
        (
            'A data bus error has occurred, and the PC value stacked for',
            'the exception return points to the instruction that caused',
            'the fault. When the processor sets this bit to 1, it writes',
            'the faulting address to the BFAR',
        ),
    ),
    BitField(
        'IMPRECISERR',
        PW_CORTEX_M_CFSR_IMPRECISERR_MASK,
        'Imprecise data bus error.',
        (
            'A data bus error has occurred, but the return address in the',
            'stack frame is not related to the instruction that caused the',
            'error. This is an asynchronous fault. Therefore, if it is',
            'detected when the priority of the current processes is higher',
            'than the BusFault priority, the BusFault becomes pending and',
            'becomes active only when the processor returns from all higher',
            'priority processes. If a precise fault occurs before the',
            'processor enters the handler for the imprecise BusFault, the',
            'handler detects both IMPRECISERR set to 1 and one of the',
            'precise fault status bits set to 1',
        ),
    ),
    BitField(
        'UNSTKERR',
        PW_CORTEX_M_CFSR_UNSTKERR_MASK,
        'BusFault on Unstacking for a return from exception.',
        (
            'Unstack for an exception return has caused one or more',
            'BusFaults. This fault is chained to the handler. This means',
            'when the processor sets this bit to 1, the original return',
            'stack is still present. The processor does not adjust the SP',
            'from the failing return, does not perform a new save, and does',
            'not write a fault address to the BFAR',
        ),
    ),
    BitField(
        'STKERR',
        PW_CORTEX_M_CFSR_STKERR_MASK,
        'BusFault on Stacking for Exception Entry.',
        (
            'Stacking for an exception entry has caused one or more',
            'BusFaults. When the processor sets this bit to 1, the SP is',
            'still adjusted but the values in the context area on the stack',
            'might be incorrect. The processor does not write a fault',
            'address to the BFAR',
        ),
    ),
    BitField(
        'LSPERR',
        PW_CORTEX_M_CFSR_LSPERR_MASK,
        'BusFault during FPU lazy state preservation.',
        '',
    ),
    BitField(
        'BFARVALID', PW_CORTEX_M_CFSR_BFARVALID_MASK, 'BFAR is valid.', ''
    ),
    BitField(
        'UNDEFINSTR',
        PW_CORTEX_M_CFSR_UNDEFINSTR_MASK,
        'Undefined Instruction UsageFault.',
        (
            'The processor has attempted to execute an undefined',
            'instruction. When this bit is set to 1, the PC value stacked',
            'for the exception return points to the undefined instruction.',
            'An undefined instruction is an instruction that the processor',
            'cannot decode.',
        ),
    ),
    BitField(
        'INVSTATE',
        PW_CORTEX_M_CFSR_INVSTATE_MASK,
        'Invalid State UsageFault.',
        (
            'The processor has attempted to execute an instruction that',
            'makes illegal use of the EPSR. The PC value stacked for the',
            'exception return points to the instruction that attempt',
            'illegal use of the EPSR',
        ),
    ),
    BitField(
        'INVPC',
        PW_CORTEX_M_CFSR_INVPC_MASK,
        'Invalid PC Load UsageFault.',
        (
            'The processor has attempted an illegal load of EXC_RETURN',
            'to the PC, as a result of an invalid context, or an invalid',
            'EXC_RETURN value. The PC value stacked for the exception',
            'return points to the instruction that tried to perform the',
            'illegal load of the PC',
        ),
    ),
    BitField(
        'NOCP',
        PW_CORTEX_M_CFSR_NOCP_MASK,
        'Coprocessor disabled or not present.',
        '',
    ),
    BitField('STKOF', PW_CORTEX_M_CFSR_STKOF_MASK, 'Stack overflowed.', ''),
    BitField(
        'UNALIGNED',
        PW_CORTEX_M_CFSR_UNALIGNED_MASK,
        'Unaligned access UsageFault.',
        (
            'The processor has made an unaligned memory access. This fault',
            'can be enabled or disabled using the UNALIGN_TRP bit in the',
            'CCR. Unaligned LDM, STM, LDRD, and STRD instructions always',
            'fault irrespective of the CCR setting.',
        ),
    ),
    BitField(
        'DIVBYZERO',
        PW_CORTEX_M_CFSR_DIVBYZERO_MASK,
        'Divide by zero.',
        (
            'The processor has executed an SDIV or UDIV instruction with',
            'a divisor of 0. The PC value stacked for the exception',
            'return points to the instruction that performed the divide',
            'by zero. This fault can be enabled or disabled using the',
            'DIV_0_TRP bit in the CCR.',
        ),
    ),
]
